from __future__ import annotations

import itertools
from typing import TYPE_CHECKING, Optional

import pytest

from qcarchivetesting import load_molecule_data
from qcfractal.components.neb.record_db_models import NEBRecordORM
from qcfractal.components.neb.testing_helpers import (
    compare_neb_specs,
    test_specs,
    submit_procedure_data,
    run_procedure_data,
)
from qcportal.neb import (
    NEBSpecification,
    NEBKeywords,
)
from qcportal.record_models import RecordStatusEnum, PriorityEnum
from qcportal.singlepoint import QCSpecification
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from qcarchivetesting.testing_classes import QCATestingSnowflake
    from qcportal import PortalClient


def test_neb_client_tag_priority(snowflake_client: PortalClient):
    chain = [load_molecule_data("neb/neb_HCN_%i" % i) for i in range(11)]

    for tag, priority in itertools.product(["*", "tag99"], list(PriorityEnum)):
        meta1, id1 = snowflake_client.add_nebs(
            [chain],
            "geometric",
            QCSpecification(
                program="psi4",
                method="hf",
                basis="sto-3g",
                driver="gradient",
                keywords={"tag_priority": [tag, priority]},
            ),
            None,
            NEBKeywords(),
            compute_priority=priority,
            compute_tag=tag,
        )
        assert meta1.n_inserted == 1
        rec = snowflake_client.get_records(id1, include=["service"])
        assert rec[0].service.compute_tag == tag
        assert rec[0].service.compute_priority == priority


@pytest.mark.parametrize("spec", test_specs)
def test_neb_client_add_get(submitter_client: PortalClient, spec: NEBSpecification):
    chain1 = [load_molecule_data("neb/neb_HCN_%i" % i) for i in range(11)]
    chain2 = [load_molecule_data("neb/neb_C3H2N_%i" % i) for i in range(21)]

    time_0 = now_at_utc()
    meta, id = submitter_client.add_nebs(
        initial_chains=[chain1, chain2],
        program=spec.program,
        singlepoint_specification=spec.singlepoint_specification,
        optimization_specification=None,
        keywords=spec.keywords,
        compute_tag="tag1",
        compute_priority=PriorityEnum.low,
    )

    time_1 = now_at_utc()
    assert meta.success

    recs = submitter_client.get_nebs(id, include=["service", "initial_chain"])

    assert len(recs) == 2

    for r in recs:
        assert r.record_type == "neb"
        assert r.record_type == "neb"
        assert compare_neb_specs(spec, r.specification)

        assert r.status == RecordStatusEnum.waiting
        assert r.children_status == {}

        assert r.service.compute_tag == "tag1"
        assert r.service.compute_priority == PriorityEnum.low

        assert r.creator_user == submitter_client.username

        assert time_0 < r.created_on < time_1
        assert time_0 < r.modified_on < time_1

    assert len(recs[0].initial_chain) == 11  # default image number
    assert len(recs[1].initial_chain) == 11

    assert recs[0].initial_chain[0].get_hash() == chain1[0].get_hash()

    hash1 = recs[0].initial_chain[0].get_hash()
    hash2 = recs[1].initial_chain[-1].get_hash()
    assert {hash1, hash2} == {chain1[0].get_hash(), chain2[-1].get_hash()}


@pytest.mark.parametrize("spec", test_specs)
@pytest.mark.parametrize("find_existing", [True, False])
def test_neb_client_add_duplicate(submitter_client: PortalClient, spec: NEBSpecification, find_existing: bool):
    chain1 = [load_molecule_data("neb/neb_HCN_%i" % i) for i in range(11)]
    chain2 = [load_molecule_data("neb/neb_C3H2N_%i" % i) for i in range(21)]
    all_chains = [chain1, chain2]

    meta, id = submitter_client.add_nebs(
        initial_chains=all_chains,
        program=spec.program,
        singlepoint_specification=spec.singlepoint_specification,
        optimization_specification=None,
        keywords=spec.keywords,
        compute_tag="tag1",
        compute_priority=PriorityEnum.low,
        find_existing=True,
    )

    assert meta.success
    print(meta)
    assert meta.n_inserted == len(all_chains)

    meta, id2 = submitter_client.add_nebs(
        initial_chains=all_chains,
        program=spec.program,
        singlepoint_specification=spec.singlepoint_specification,
        optimization_specification=None,
        keywords=spec.keywords,
        compute_tag="tag1",
        compute_priority=PriorityEnum.low,
        find_existing=find_existing,
    )

    if find_existing:
        assert meta.n_existing == len(all_chains)
        assert meta.n_inserted == 0
        assert id == id2
    else:
        assert meta.n_existing == 0
        assert meta.n_inserted == len(all_chains)
        assert set(id).isdisjoint(id2)


def test_neb_client_add_existing_chain(snowflake_client: PortalClient):
    spec = test_specs[0]
    chain1 = [load_molecule_data("neb/neb_HCN_%i" % i) for i in range(11)]
    chain2 = [load_molecule_data("neb/neb_C3H2N_%i" % i) for i in range(21)]

    # Add a chain separately
    _, mol_ids = snowflake_client.add_molecules(chain1)

    # Now add records
    meta, id = snowflake_client.add_nebs(
        initial_chains=[chain1, chain2],
        program=spec.program,
        keywords=spec.keywords,
        singlepoint_specification=spec.singlepoint_specification,
        optimization_specification=None,
        compute_tag="tag1",
        compute_priority=PriorityEnum.low,
    )
    recs = snowflake_client.get_nebs(id)

    assert len(recs) == 2
    assert recs[0].initial_chain[0].id == mol_ids[0]


def test_neb_client_delete(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    neb_id = run_procedure_data(storage_socket, activated_manager_name, "neb_HCN_psi4_pbe_opt2")

    with storage_socket.session_scope() as session:
        rec = session.get(NEBRecordORM, neb_id)

        # Children are singlepoints, optimizations, and the trajectory of the optimizations (also singlepoints)
        direct_child_ids = [x.singlepoint_id for x in rec.singlepoints]
        opt_ids = [x.optimization_id for x in rec.optimizations]
        direct_child_ids.extend(opt_ids)

        child_ids = direct_child_ids.copy()
        for opt in rec.optimizations:
            traj_ids = [x.singlepoint_id for x in opt.optimization_record.trajectory]
            child_ids.extend(traj_ids)

    # Some duplicates here
    direct_child_ids = list(set(direct_child_ids))
    child_ids = list(set(child_ids))

    meta = snowflake_client.delete_records(neb_id, soft_delete=True, delete_children=False)
    assert meta.success
    assert meta.deleted_idx == [0]
    assert meta.n_children_deleted == 0

    child_recs = snowflake_client.get_records(child_ids, missing_ok=True)
    assert all(x.status == RecordStatusEnum.complete for x in child_recs)
    neb_rec = snowflake_client.get_records(neb_id)
    assert neb_rec.children_status == {RecordStatusEnum.complete: len(direct_child_ids)}

    snowflake_client.undelete_records(neb_id)

    meta = snowflake_client.delete_records(neb_id, soft_delete=True, delete_children=True)
    assert meta.success
    assert meta.deleted_idx == [0]
    assert meta.n_children_deleted == len(child_ids)

    child_recs = snowflake_client.get_records(child_ids, missing_ok=True)
    assert all(x.status == RecordStatusEnum.deleted for x in child_recs)
    neb_rec = snowflake_client.get_records(neb_id)
    assert neb_rec.children_status == {RecordStatusEnum.deleted: len(direct_child_ids)}

    meta = snowflake_client.delete_records(neb_id, soft_delete=False, delete_children=True)
    assert meta.success
    assert meta.deleted_idx == [0]
    assert meta.n_children_deleted == len(child_ids)

    recs = snowflake_client.get_nebs(neb_id, missing_ok=True)
    assert recs is None

    child_recs = snowflake_client.get_records(child_ids, missing_ok=True)
    assert all(x is None for x in child_recs)

    # DB should be pretty empty now
    query_res = snowflake_client.query_records()
    query_res_l = list(query_res)
    assert len(query_res_l) == 0


def test_neb_client_harddelete_nochildren(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    neb_id = run_procedure_data(storage_socket, activated_manager_name, "neb_HCN_psi4_pbe_opt2")

    with storage_socket.session_scope() as session:
        rec = session.get(NEBRecordORM, neb_id)
        child_ids = [x.singlepoint_id for x in rec.singlepoints]

    meta = snowflake_client.delete_records(neb_id, soft_delete=False, delete_children=False)
    assert meta.success
    assert meta.deleted_idx == [0]
    assert meta.n_children_deleted == 0

    for cid in child_ids:
        child_rec = snowflake_client.get_records(cid, missing_ok=True)
        assert child_rec is not None


def test_neb_client_delete_opt_inuse(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    neb_id = run_procedure_data(storage_socket, activated_manager_name, "neb_HCN_psi4_pbe_opt2")

    with storage_socket.session_scope() as session:
        rec = session.get(NEBRecordORM, neb_id)
        child_ids = [x.singlepoint_id for x in rec.singlepoints]

    meta = snowflake_client.delete_records(child_ids[0], soft_delete=False)
    assert meta.success is False
    assert meta.error_idx == [0]

    ch_rec = snowflake_client.get_records(child_ids[0])
    assert ch_rec is not None


def test_neb_client_query(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    id_1, _ = submit_procedure_data(storage_socket, "neb_HCN_psi4_pbe")
    id_2, _ = submit_procedure_data(storage_socket, "neb_HCN_psi4_pbe0_opt1")
    id_3, _ = submit_procedure_data(storage_socket, "neb_HCN_psi4_pbe_opt2")
    id_4, _ = submit_procedure_data(storage_socket, "neb_HCN_psi4_b3lyp_opt3")

    all_records = snowflake_client.get_nebs([id_1, id_2, id_3, id_4], include=["initial_chain"])

    # mol_ids of just first chain (11 images, the other three have 7 images).
    mol_ids = [x.id for x in all_records[1].initial_chain]

    query_res = snowflake_client.query_nebs(qc_program=["psi4"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 4

    query_res = snowflake_client.query_nebs(qc_program=["nothing"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    query_res = snowflake_client.query_nebs(molecule_id=mol_ids[1])
    query_res_l = list(query_res)
    assert len(query_res_l) == 4

    # Multiple molecules that belong to the same NEB calculation
    query_res = snowflake_client.query_nebs(molecule_id=mol_ids[1:4])
    query_res_l = list(query_res)
    assert len(query_res_l) == 4

    query_res = snowflake_client.query_nebs(molecule_id=[mol_ids[10]])
    query_res_l = list(query_res)
    assert len(query_res_l) == 4

    query_res = snowflake_client.query_nebs(molecule_id=999999999)
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    query_res = snowflake_client.query_nebs(program=["geometric"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 4

    query_res = snowflake_client.query_nebs(program=["geometric123"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    query_res = snowflake_client.query_nebs(qc_basis=["6-31g"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 1

    query_res = snowflake_client.query_nebs(qc_basis=[None])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    query_res = snowflake_client.query_nebs(qc_basis=[""])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    # query for method
    query_res = snowflake_client.query_nebs(qc_method=["b3lyP"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 1

    # Query by default returns everything
    query_res = snowflake_client.query_nebs()
    query_res_l = list(query_res)
    assert len(query_res_l) == 4

    # Query by default (with a limit)
    query_res = snowflake_client.query_nebs(limit=1)
    query_res_l = list(query_res)
    assert len(query_res_l) == 1



from flask import jsonify, current_app, g

from qcfractal.flask_app.api_v1.blueprint import api_v1


@api_v1.route("/ping", methods=["GET"])
def ping():

    user_id = g.user_id if "user_id" in g else None

    allow_unauth_read = current_app.config["QCFRACTAL_CONFIG"].allow_unauthenticated_read
    is_authorized = True if (user_id is not None or allow_unauth_read) else False
    ret = {"success": True, "authorized": is_authorized}

    if user_id is not None:
        ret["user_info"] = {
            "user_id": user_id,
            "username": g.username,
            "role": g.role,
            "groups": g.groups,
        }

    return jsonify(ret)



import time

import jwt
import pytest

from qcarchivetesting import test_users
from qcarchivetesting.testing_classes import (
    QCATestingSnowflake,
)
from qcportal.exceptions import AuthenticationFailure


@pytest.mark.slow
def test_jwt_refresh(secure_snowflake):
    client = secure_snowflake.client("submit_user", password=test_users["submit_user"]["pw"])
    time.sleep(client._jwt_access_exp - time.time() + 1)
    client.list_datasets()


@pytest.mark.slow
def test_jwt_refresh_user_newrole(secure_snowflake):
    admin_client = secure_snowflake.client("admin_user", password=test_users["admin_user"]["pw"])
    client = secure_snowflake.client("submit_user", password=test_users["submit_user"]["pw"])

    submit_info = client.get_user()
    assert submit_info.role == "submit"
    decoded_access = jwt.decode(client._jwt_access_token, algorithms=["HS256"], options={"verify_signature": False})
    assert decoded_access["role"] == "submit"

    uinfo = admin_client.get_user("submit_user")
    uinfo.role = "read"
    admin_client.modify_user(uinfo)

    time.sleep(client._jwt_access_exp - time.time() + 1)

    client.list_datasets()  # will refresh token as needed
    decoded_access = jwt.decode(client._jwt_access_token, algorithms=["HS256"], options={"verify_signature": False})
    assert decoded_access["role"] == "read"


@pytest.mark.slow
def test_jwt_refresh_user_disabled(secure_snowflake):
    admin_client = secure_snowflake.client("admin_user", password=test_users["admin_user"]["pw"])
    client = secure_snowflake.client("submit_user", password=test_users["submit_user"]["pw"])

    uinfo = admin_client.get_user("submit_user")
    uinfo.enabled = False
    admin_client.modify_user(uinfo)

    time.sleep(client._jwt_access_exp - time.time() + 1)

    with pytest.raises(AuthenticationFailure, match="User account has been disabled"):
        client.list_datasets()


@pytest.mark.slow
def test_jwt_refresh_user_deleted(postgres_server, pytestconfig):
    # Need its own snowflake because we need logging disabled
    # Otherwise, the user cannot be deleted because it is referenced in the access log table

    pg_harness = postgres_server.get_new_harness("jwt_user_deleted")
    encoding = pytestconfig.getoption("--client-encoding")
    with QCATestingSnowflake(
        pg_harness,
        encoding=encoding,
        create_users=True,
        enable_security=True,
        allow_unauthenticated_read=False,
        log_access=False,
    ) as snowflake:
        admin_client = snowflake.client("admin_user", password=test_users["admin_user"]["pw"])
        client = snowflake.client("submit_user", password=test_users["submit_user"]["pw"])

        admin_client.delete_user("submit_user")
        time.sleep(client._jwt_access_exp - time.time() + 1)

        with pytest.raises(AuthenticationFailure, match="User account no longer exists"):
            client.list_datasets()



"""Gridoptimization spec hashing

Revision ID: 12d37bad13ba
Revises: 518ee94f4749
Create Date: 2024-12-17 15:41:38.740543

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

from qcfractal.alembic.migration_helpers.hashing import hash_dict_1

# revision identifiers, used by Alembic.
revision = "12d37bad13ba"
down_revision = "518ee94f4749"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "gridoptimization_specification", sa.Column("protocols", postgresql.JSONB(astext_type=sa.Text()), nullable=True)
    )
    op.add_column("gridoptimization_specification", sa.Column("specification_hash", sa.String(), nullable=True))

    # Generate hashes for existing specs
    conn = op.get_bind()
    res = conn.execute(sa.text(f"SELECT id, program, keywords FROM gridoptimization_specification;"))
    all_spec = res.fetchall()
    for spec_id, program, keywords in all_spec:
        d = {
            "program": program,
            "keywords": keywords,
            "protocols": {},  # Default for now
        }
        h = hash_dict_1(d)
        op.execute(
            sa.text(f"""UPDATE gridoptimization_specification SET specification_hash = '{h}' WHERE id = {spec_id};""")
        )

    op.execute(sa.text("UPDATE gridoptimization_specification SET protocols = '{}'::JSONB"))

    op.alter_column("gridoptimization_specification", "specification_hash", nullable=False)
    op.alter_column("gridoptimization_specification", "protocols", nullable=False)

    op.drop_constraint("ux_gridoptimization_specification_keys", "gridoptimization_specification", type_="unique")
    op.drop_column("gridoptimization_specification", "keywords_hash")
    op.create_unique_constraint(
        "ux_gridoptimization_specification_keys",
        "gridoptimization_specification",
        ["specification_hash", "optimization_specification_id"],
    )
    # ### end Alembic commands ###


def downgrade():
    raise RuntimeError("Downgrade not supported.")



"""Remove created_on from task/service queue

Revision ID: d1ee87a66b71
Revises:
Create Date: 2023-08-15 17:58:28.061638

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "d1ee87a66b71"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_service_queue_waiting_sort", table_name="service_queue")
    op.drop_column("service_queue", "created_on")
    op.drop_index("ix_task_queue_waiting_sort", table_name="task_queue")
    op.drop_column("task_queue", "created_on")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("task_queue", sa.Column("created_on", postgresql.TIMESTAMP(), autoincrement=False, nullable=False))
    op.create_index("ix_task_queue_waiting_sort", "task_queue", [sa.text("priority DESC"), "created_on"], unique=False)
    op.add_column("service_queue", sa.Column("created_on", postgresql.TIMESTAMP(), autoincrement=False, nullable=False))
    op.create_index(
        "ix_service_queue_waiting_sort", "service_queue", [sa.text("priority DESC"), "created_on"], unique=False
    )
    # ### end Alembic commands ###



from __future__ import annotations

import logging
import secrets
from typing import TYPE_CHECKING

import bcrypt
from sqlalchemy.exc import IntegrityError
from sqlalchemy.sql import select

from qcportal.auth import UserInfo, is_valid_password, is_valid_username, AuthTypeEnum
from qcportal.exceptions import AuthenticationFailure, UserManagementError, InvalidRolenameError
from .db_models import UserORM, UserGroupORM, UserPreferencesORM
from .role_permissions import GLOBAL_ROLE_PERMISSIONS

if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session
    from qcfractal.db_socket.socket import SQLAlchemySocket
    from typing import Optional, List, Dict, Any, Union, Tuple, Set


valid_roles: Set[str] = set(GLOBAL_ROLE_PERMISSIONS.keys())


def is_valid_role(role: str):
    if role not in valid_roles:
        raise InvalidRolenameError(f"Role with name '{role} does not exist")


def _generate_password() -> str:
    """
    Generates a random password

    Returns
    -------
    :
        An plain-text random password.
    """
    return secrets.token_urlsafe(16)


def _hash_password(password: str) -> bytes:
    """
    Hashes a password in a consistent way
    """

    return bcrypt.hashpw(password.encode("UTF-8"), bcrypt.gensalt(6))


class UserSocket:
    """
    Socket for managing users
    """

    def __init__(self, root_socket: SQLAlchemySocket):
        self.root_socket = root_socket
        self._logger = logging.getLogger(__name__)

    def _get_internal(self, session: Session, username_or_id: Union[int, str]) -> UserORM:
        """
        Obtain the ORM for a particular user.

        If the user is not found, an exception is raised. The ORM is attached to the given session

        Parameters
        ----------
        session
            SQLAlchemy session to use for querying
        username_or_id
            The username or user ID

        Returns
        -------
        :
            ORM of the specified user
        """

        if isinstance(username_or_id, int) or username_or_id.isdecimal():
            stmt = select(UserORM).where(UserORM.id == username_or_id)
        else:
            is_valid_username(username_or_id)
            stmt = select(UserORM).where(UserORM.username == username_or_id)

        user = session.execute(stmt).scalar_one_or_none()

        if user is None:
            raise UserManagementError(f"User {username_or_id} not found.")

        return user

    def _assert_user_exists(self, session: Session, username_or_id: Union[int, str]) -> None:
        # Just call the existing function, swallowing the return
        _ = self._get_internal(session, username_or_id)

    def list(self, *, session: Optional[Session] = None) -> List[Dict[str, Any]]:
        """
        Get information about all users

        Parameters
        ----------
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        with self.root_socket.optional_session(session, True) as session:
            stmt = select(UserORM).order_by(UserORM.id.asc())
            all_users = session.execute(stmt).scalars().all()
            return [x.model_dict() for x in all_users]

    def get(self, username_or_id: Union[int, str], *, session: Optional[Session] = None) -> Dict[str, Any]:
        """
        Obtains information for a user

        Returns all info for a user, except (hashed) password

        Parameters
        ----------
        username_or_id
            The username or user ID
        session
            An existing SQLAlchemy session to use. If None, one will be created
        """

        with self.root_socket.optional_session(session, True) as session:
            user = self._get_internal(session, username_or_id)
            return user.model_dict()

    def add(self, user_info: UserInfo, password: Optional[str] = None, *, session: Optional[Session] = None) -> str:
        """
        Adds a new user

        Parameters
        ----------
        user_info
            New user's information
        password
            The user's password. If None, a new password will be generated.
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            The password for the user. This is useful if the password is autogenerated
        """

        # Should have been checked already, but defense in depth
        is_valid_username(user_info.username)
        is_valid_role(user_info.role)

        # ID should not be set
        if user_info.id is not None:
            raise UserManagementError("Cannot add a user - id was given as part of new user info")

        if password is None:
            password = _generate_password()

        is_valid_password(password)

        hashed_pw = _hash_password(password)

        # Groups are not directly a part of the ORM
        user_dict = user_info.dict(exclude={"groups"})

        try:
            with self.root_socket.optional_session(session) as session:
                # Will raise exception if group does not exist or name is invalid
                groups = [self.root_socket.groups._get_internal(session, g) for g in user_info.groups]

                user = UserORM(**user_dict, groups_orm=groups, password=hashed_pw)
                session.add(user)
        except IntegrityError:
            raise UserManagementError(f"User {user_info.username} already exists")

        self._logger.info(f"User {user_info.username} added")
        return password

    def _verify_local_password(self, user: UserORM, password: str):
        """
        Verifies a given username and password against the local db

        Raises exception if the password does not match or there is another problem
        """

        is_valid_password(password)

        try:
            pwcheck = bcrypt.checkpw(password.encode("UTF-8"), user.password)
        except Exception as e:
            self._logger.error(f"Password check failure for user {user.username}, error: {str(e)}")
            self._logger.error(
                f"Error likely caused by encryption salt mismatch, potentially fixed by creating a new password for user {user.username}."
            )
            raise UserManagementError("Password decryption failure, please contact your system administrator.")

        if pwcheck is False:
            raise AuthenticationFailure("Incorrect username or password")

    def authenticate(self, username: str, password: str, *, session: Optional[Session] = None) -> UserInfo:
        """
        Authenticates a given username and password, returning all info about the user

        If the user is not found, or is disabled, or the password is incorrect, an exception is raised.

        Parameters
        ----------
        username
            The username of the user
        password
            The password associated with the username
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        --------
        :
            All information about the user
        """

        is_valid_username(username)

        with self.root_socket.optional_session(session, True) as session:
            try:
                user = self._get_internal(session, username)
            except UserManagementError as e:
                # Turn missing user into an Authentication error
                raise AuthenticationFailure("Incorrect username or password")

            if not user.enabled:
                raise AuthenticationFailure(f"User {username} is disabled.")

            # what's next depends on how the user is authenticated
            if user.auth_type == AuthTypeEnum.password:
                self._verify_local_password(user=user, password=password)
            else:
                self._logger.error(f"Unknown auth type: {user.auth_type}. This is a developer error")
                raise UserManagementError(f"Unknown authentication type stored in the database: {user.auth_type}")

            return user.to_model(UserInfo)

    def modify(self, user_info: UserInfo, as_admin: bool, *, session: Optional[Session] = None) -> Dict[str, Any]:
        """
        Alters a user's information

        The user to modify is taken from the user_info object.

        The user's username or password cannot be changed this way. If `as_admin` is False, then only
        the descriptive changes (email, etc) can be changed. If it is True, then
        the `enabled` and `role` fields can also be changed.

        Parameters
        ----------
        user_info
            The user info to update the database with
        as_admin
            Enable changing sensitive columns (enabled & role)
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            An updated version of the user info, with all possible/allowed changes

        """

        with self.root_socket.optional_session(session) as session:
            user = self._get_internal(session, user_info.id)

            if user_info.username != user.username:
                raise UserManagementError("Cannot change username")

            user.fullname = user_info.fullname
            user.organization = user_info.organization
            user.email = user_info.email

            if as_admin is True:
                is_valid_role(user_info.role)
                groups = [self.root_socket.groups._get_internal(session, g) for g in user_info.groups]

                user.enabled = user_info.enabled
                user.role = user_info.role
                user.groups_orm = groups

            session.commit()

            self._logger.info(f"User {user_info.username} modified")

            return self.get(user_info.username, session=session)

    def change_password(
        self, username_or_id: Union[int, str], password: Optional[str], *, session: Optional[Session] = None
    ) -> str:
        """
        Alters a user's password

        Parameters
        ----------
        username_or_id
            The username or ID of the user
        password
            The user's new password. If the password is empty, an exception is raised. If None, then a
            password will be generated
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            A string representing the password. If a new password was given, this should be identical
            to the input password. Otherwise, it will be the generated password.
        """

        if password is None:
            password = _generate_password()

        is_valid_password(password)

        with self.root_socket.optional_session(session) as session:
            user = self._get_internal(session, username_or_id)
            user.password = _hash_password(password)

        self._logger.info(f"Password for {username_or_id} modified")
        return password

    def delete(self, username_or_id: Union[int, str], *, session: Optional[Session] = None) -> None:
        """Removes a user

        This will raise an exception if the user doesn't exist or is being referenced elsewhere in the
        database.

        Parameters
        ----------
        username_or_id
            The username or ID of the user
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        try:
            with self.root_socket.optional_session(session) as session:
                user = self._get_internal(session, username_or_id)
                session.delete(user)
        except IntegrityError:
            raise UserManagementError("User could not be deleted. Likely it is being referenced somewhere")

        self._logger.info(f"User {username_or_id} deleted")

    def get_optional_user_id(
        self,
        username_or_id: Optional[Union[int, str]],
        *,
        session: Optional[Session] = None,
    ) -> Optional[int]:
        """
        Obtain the ID of a user

        If username_or_id is None, None is returned.

        If an ID or name that does not exist is given, an exception is raised.
        """

        if username_or_id is None:
            return None

        with self.root_socket.optional_session(session) as session:
            user = self._get_internal(session, username_or_id)
            return user.id

    def get_preferences(self, user_id: int, *, session: Optional[Session] = None) -> Dict[str, Any]:
        """
        Get the user-set preferences for a given user

        Raises an exception if the user is not found
        """

        stmt = select(UserPreferencesORM).where(UserPreferencesORM.user_id == user_id)

        with self.root_socket.optional_session(session, True) as session:
            self._assert_user_exists(session, user_id)

            r = session.execute(stmt).scalar_one_or_none()
            if r is None:
                return {}

            return r.preferences

    def set_preferences(self, user_id: int, preferences: Dict[str, Any], *, session: Optional[Session] = None) -> None:
        """
        Sets a users preferences for a given user
        """

        with self.root_socket.optional_session(session) as session:
            self._assert_user_exists(session, user_id)

            stmt = select(UserPreferencesORM).where(UserPreferencesORM.user_id == user_id)
            prefs_orm = session.execute(stmt).scalar_one_or_none()
            if prefs_orm is None:
                prefs_orm = UserPreferencesORM(user_id=user_id, preferences=preferences)
                session.add(prefs_orm)
            else:
                prefs_orm.preferences = preferences



from __future__ import annotations

from typing import TYPE_CHECKING

from .flask_app import create_flask_app

if TYPE_CHECKING:
    import queue
    from typing import Optional
    from ..config import FractalConfig


class FractalWaitressApp:
    def __init__(
        self,
        qcfractal_config: FractalConfig,
        finished_queue: Optional[queue.Queue] = None,
    ):
        self.qcfractal_config = qcfractal_config
        self.application = create_flask_app(qcfractal_config, finished_queue=finished_queue)

    def run(self):
        from waitress import serve

        waitress_opts = self.qcfractal_config.api.extra_waitress_options
        if waitress_opts is None:
            waitress_opts = {}

        serve(
            self.application,
            host=self.qcfractal_config.api.host,
            port=self.qcfractal_config.api.port,
            threads=self.qcfractal_config.api.num_threads_per_worker,
            **waitress_opts,
        )



from qcarchivetesting.helpers import load_hash_test_data
from qcfractal.components.gridoptimization.record_db_models import GridoptimizationSpecificationORM
from qcfractal.components.testing_fixtures import spec_test_runner
from qcfractal.db_socket import SQLAlchemySocket
from qcportal.gridoptimization import GridoptimizationSpecification, GridoptimizationKeywords
from qcportal.optimization import OptimizationSpecification, OptimizationProtocols
from qcportal.singlepoint import QCSpecification, SinglepointDriver, SinglepointProtocols

common_opt_spec = OptimizationSpecification(
    program="optprog2",
    keywords={"k": "value"},
    protocols=OptimizationProtocols(trajectory="none"),
    qc_specification=QCSpecification(
        program="prog2",
        driver=SinglepointDriver.hessian,
        method="hf",
        basis="def2-tzvp",
        keywords={"k": "value"},
        protocols=SinglepointProtocols(wavefunction="orbitals_and_eigenvalues"),
    ),
)


def test_gridoptimization_hash_canaries(storage_socket: SQLAlchemySocket):
    # Test data is hash : spec dict
    test_data = load_hash_test_data("gridoptimization_specification_tests")

    # Hashes are independent of opt spec
    # TODO - protocols are not part of model
    for t in test_data.values():
        del t["protocols"]
    spec_map = [
        (k, GridoptimizationSpecification(optimization_specification=common_opt_spec, **v))
        for k, v in test_data.items()
    ]

    specs = [x[1] for x in spec_map]
    meta, ids = storage_socket.records.gridoptimization.add_specifications(specs)
    assert meta.success
    assert len(ids) == len(specs)
    assert meta.n_existing == 0

    with storage_socket.session_scope() as session:
        for spec_id, (spec_hash, _) in zip(ids, spec_map):
            spec_orm = session.get(GridoptimizationSpecificationORM, spec_id)
            assert spec_orm.specification_hash == spec_hash


def test_gridoptimization_socket_add_specification_same_1(spec_test_runner):
    spec1 = GridoptimizationSpecification(
        program="gridoptimization",
        keywords=GridoptimizationKeywords(
            preoptimization=False,
            scans=[
                {"type": "distance", "indices": [0, 3], "steps": [2.0, 3.0, 4.0], "step_type": "relative"},
            ],
        ),
        optimization_specification=common_opt_spec,
    )

    spec_test_runner("gridoptimization", spec1, spec1, True)


def test_gridoptimization_socket_add_specification_same_2(spec_test_runner):
    # capitalization should be ignored
    spec1 = GridoptimizationSpecification(
        program="gridoptimization",
        keywords=GridoptimizationKeywords(
            preoptimization=False,
            scans=[
                {"type": "distance", "indices": [0, 3], "steps": [2.0, 3.0, 4.0], "step_type": "relative"},
            ],
        ),
        optimization_specification=common_opt_spec,
    )

    spec2 = GridoptimizationSpecification(
        program="gridopTIMIZatiON",
        keywords=GridoptimizationKeywords(
            preoptimization=False,
            scans=[
                {"type": "distance", "indices": [0, 3], "steps": [2.0, 3.0, 4.0], "step_type": "relative"},
            ],
        ),
        optimization_specification=common_opt_spec,
    )

    spec_test_runner("gridoptimization", spec1, spec2, True)


def test_gridoptimization_socket_add_specification_diff_1(spec_test_runner):
    # different indices
    spec1 = GridoptimizationSpecification(
        program="gridoptimization",
        keywords=GridoptimizationKeywords(
            preoptimization=False,
            scans=[
                {"type": "distance", "indices": [0, 3], "steps": [2.0, 3.0, 4.0], "step_type": "relative"},
            ],
        ),
        optimization_specification=common_opt_spec,
    )

    go_kw = spec1.keywords.copy(
        update={"scans": [{"type": "distance", "indices": [0, 4], "steps": [2.0, 3.0, 4.0], "step_type": "relative"}]}
    )
    spec2 = spec1.copy(update={"keywords": go_kw})
    spec_test_runner("gridoptimization", spec1, spec2, False)


def test_gridoptimization_socket_add_specification_diff_2(spec_test_runner):
    # different steps
    spec1 = GridoptimizationSpecification(
        program="gridoptimization",
        keywords=GridoptimizationKeywords(
            preoptimization=False,
            scans=[
                {"type": "distance", "indices": [0, 3], "steps": [2.0, 3.0, 4.0], "step_type": "relative"},
            ],
        ),
        optimization_specification=common_opt_spec,
    )

    go_kw = spec1.keywords.copy(
        update={"scans": [{"type": "distance", "indices": [0, 3], "steps": [2.0, 3.0, 4.1], "step_type": "relative"}]}
    )
    spec2 = spec1.copy(update={"keywords": go_kw})
    spec_test_runner("gridoptimization", spec1, spec2, False)


def test_gridoptimization_socket_add_specification_diff_3(spec_test_runner):
    # absolute vs relative
    spec1 = GridoptimizationSpecification(
        program="gridoptimization",
        keywords=GridoptimizationKeywords(
            preoptimization=False,
            scans=[
                {"type": "distance", "indices": [0, 3], "steps": [2.0, 3.0, 4.0], "step_type": "relative"},
            ],
        ),
        optimization_specification=common_opt_spec,
    )

    go_kw = spec1.keywords.copy(
        update={"scans": [{"type": "distance", "indices": [0, 3], "steps": [2.0, 3.0, 4.0], "step_type": "absolute"}]}
    )
    spec2 = spec1.copy(update={"keywords": go_kw})
    spec_test_runner("gridoptimization", spec1, spec2, False)


def test_gridoptimization_socket_add_specification_diff_4(spec_test_runner):
    # preoptimization
    spec1 = GridoptimizationSpecification(
        program="gridoptimization",
        keywords=GridoptimizationKeywords(
            preoptimization=False,
            scans=[
                {"type": "distance", "indices": [0, 3], "steps": [2.0, 3.0, 4.0], "step_type": "relative"},
            ],
        ),
        optimization_specification=common_opt_spec,
    )

    go_kw = spec1.keywords.copy(update={"preoptimization": True})
    spec2 = spec1.copy(update={"keywords": go_kw})
    spec_test_runner("gridoptimization", spec1, spec2, False)


def test_gridoptimization_socket_add_specification_diff_5(spec_test_runner):
    # basis set
    spec1 = GridoptimizationSpecification(
        program="gridoptimization",
        keywords=GridoptimizationKeywords(
            preoptimization=False,
            scans=[
                {"type": "distance", "indices": [0, 3], "steps": [2.0, 3.0, 4.0], "step_type": "relative"},
            ],
        ),
        optimization_specification=common_opt_spec,
    )

    qc_spec = spec1.optimization_specification.qc_specification.copy(update={"basis": "def2-qzvp"})
    opt_spec = spec1.optimization_specification.copy(update={"qc_specification": qc_spec})
    spec2 = spec1.copy(update={"optimization_specification": opt_spec})
    spec_test_runner("gridoptimization", spec1, spec2, False)


def test_gridoptimization_socket_add_specification_diff_6(spec_test_runner):
    # qc keywords
    spec1 = GridoptimizationSpecification(
        program="gridoptimization",
        keywords=GridoptimizationKeywords(
            preoptimization=False,
            scans=[
                {"type": "distance", "indices": [0, 3], "steps": [2.0, 3.0, 4.0], "step_type": "relative"},
            ],
        ),
        optimization_specification=common_opt_spec,
    )

    qc_spec = spec1.optimization_specification.qc_specification.copy(update={"keywords": {"z": 1.0 - 10}})
    opt_spec = spec1.optimization_specification.copy(update={"qc_specification": qc_spec})
    spec2 = spec1.copy(update={"optimization_specification": opt_spec})
    spec_test_runner("gridoptimization", spec1, spec2, False)



from __future__ import annotations

import logging
import logging.handlers
import multiprocessing
import queue
import threading
import weakref
from typing import Optional

from werkzeug.serving import make_server

from qcfractal.config import FractalConfig
from qcfractal.flask_app import create_flask_app
from qcfractal.flask_app.waitress_app import FractalWaitressApp
from qcfractal.job_runner import FractalJobRunner


def api_process(
    qcf_config: FractalConfig,
    logging_queue: multiprocessing.Queue,
    finished_queue: Optional[multiprocessing.Queue],
    initialized_event: Optional[multiprocessing.Event] = None,
) -> None:
    import signal

    qh = logging.handlers.QueueHandler(logging_queue)
    logger = logging.getLogger()
    logger.handlers.clear()
    logger.addHandler(qh)
    logger.setLevel(qcf_config.loglevel)

    early_stop = False

    def signal_handler(signum, frame):
        nonlocal early_stop
        early_stop = True

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    api = FractalWaitressApp(qcf_config, finished_queue=finished_queue)

    if early_stop:
        logging_queue.close()
        logging_queue.join_thread()
        return

    def signal_handler(signum, frame):
        raise KeyboardInterrupt

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    if initialized_event is not None:
        initialized_event.set()

    try:
        api.run()
    except KeyboardInterrupt:  # Swallow ugly output on CTRL-C
        pass
    finally:
        logging_queue.close()
        logging_queue.join_thread()


class QCFAPIThread:
    """
    A class that runs the QCFractal API in a separate thread

    This does not use waitress, but instead runs the Flask app directly. This is useful for debugging and testing
    (such as with the snowflake) but shouldn't be used in user-facing production.
    """

    def __init__(
        self,
        qcf_config: FractalConfig,
        finished_queue: Optional[queue.Queue] = None,
    ):
        self._qcf_config = qcf_config
        self._flask_app = create_flask_app(qcf_config, finished_queue=finished_queue)
        self._server = make_server(self._qcf_config.api.host, self._qcf_config.api.port, self._flask_app)
        self._api_thread = None
        self._finalizer = None

    # Classmethod because finalizer can't handle bound methods
    @classmethod
    def _stop(cls, server, api_thread):
        if server is not None:
            server.shutdown()
            api_thread.join()

    def start(self, initialized_event: Optional[threading.Event] = None) -> None:
        if self._api_thread is not None:
            raise RuntimeError("API already started")

        # For some reason we need to set the level here manually? Otherwise it's always INFO
        logging.getLogger("werkzeug").setLevel(self._qcf_config.loglevel)

        # We use daemon=True
        # This means that the main python process can exit, calling various destructors
        # and finalizers (rather than waiting for those threads to finish before doing so)
        self._api_thread = threading.Thread(target=self._server.serve_forever, daemon=True)
        self._api_thread.start()

        if initialized_event is not None:
            initialized_event.set()

        self._finalizer = weakref.finalize(
            self,
            self._stop,
            self._server,
            self._api_thread,
        )

    def stop(self) -> None:
        if self._finalizer is not None:
            self._finalizer()

        self._api_thread = None

    def is_alive(self) -> bool:
        if self._api_thread is None:
            return False
        return self._api_thread.is_alive()


def job_runner_process(
    qcf_config: FractalConfig,
    logging_queue: multiprocessing.Queue,
    finished_queue: Optional[multiprocessing.Queue],
    initialized_event: Optional[multiprocessing.Event] = None,
) -> None:
    import signal

    qh = logging.handlers.QueueHandler(logging_queue)
    logger = logging.getLogger()
    logger.handlers.clear()
    logger.addHandler(qh)
    logger.setLevel(qcf_config.loglevel)

    early_stop = False

    def signal_handler(signum, frame):
        nonlocal early_stop
        early_stop = True

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    job_runner = FractalJobRunner(qcf_config, finished_queue=finished_queue)

    if early_stop:
        logging_queue.close()
        logging_queue.join_thread()
        return

    def signal_handler(signum, frame):
        job_runner.stop()

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    if initialized_event is not None:
        initialized_event.set()

    try:
        job_runner.start()
    except KeyboardInterrupt:  # Swallow ugly output on CTRL-C
        pass
    finally:
        logging_queue.close()
        logging_queue.join_thread()


class QCFJobRunnerThread:
    def __init__(
        self,
        qcf_config: FractalConfig,
        finished_queue: Optional[queue.Queue] = None,
    ):
        self._qcf_config = qcf_config
        self._job_runner: Optional[FractalJobRunner] = None
        self._job_runner_thread = None
        self._finalizer = None
        self._finished_queue = finished_queue

    # Classmethod because finalizer can't handle bound methods
    @classmethod
    def _stop(cls, job_runner, job_runner_thread):
        if job_runner is not None:
            job_runner.stop()
            job_runner_thread.join()

    def start(self, initialized_event: Optional[threading.Event] = None) -> None:
        if self._job_runner is not None:
            raise RuntimeError("Job Runner already started")

        self._job_runner = FractalJobRunner(self._qcf_config, self._finished_queue)

        # We use daemon=True
        # This means that the main python process can exit, calling various destructors
        # and finalizers (rather than waiting for those threads to finish before doing so)
        self._job_runner_thread = threading.Thread(target=self._job_runner.start, daemon=True)
        self._job_runner_thread.start()

        if initialized_event is not None:
            initialized_event.set()

        self._finalizer = weakref.finalize(
            self,
            self._stop,
            self._job_runner,
            self._job_runner_thread,
        )

    def stop(self) -> None:
        if self._finalizer is not None:
            self._finalizer()

        self._job_runner = None
        self._job_runner_thread = None

    def is_alive(self) -> bool:
        if self._job_runner_thread is None:
            return False
        return self._job_runner_thread.is_alive()



import logging

from qcfractal.components.record_db_models import BaseRecordORM
from qcfractal.config import AutoResetConfig
from qcportal.compression import decompress

# Map from specific errors to the general error classes
error_map = {
    "unknown": "unknown_error",
    "random_error": "random_error",
    "KilledWorker": "compute_lost",
    "ManagerLost": "compute_lost",
}

logger = logging.getLogger(__name__)


def should_reset(record_orm: BaseRecordORM, config: AutoResetConfig) -> bool:
    """
    Determine if a record should be automatically reset

    Parameters
    ----------
    record_orm
        ORM of the record to inspect
    config
        Configuration of the auto-reset logic

    Returns
    -------
    :
        True if the record should be reset automatically, False if it should remain errored.
    """

    history = record_orm.compute_history

    # Kinda wrote myself into a corner with all this compression stuff...
    error_orm = [x.outputs.get("error", None) for x in history]
    error_dict = [decompress(x.data, x.compression_type) for x in error_orm]

    error_types = [x["error_type"] for x in error_dict]

    # Internal error - never restart automatically
    if "internal_fractal_error" in error_types:
        return False

    # all unique error types
    unique_errors = set(error_types)

    error_counts = {x: error_types.count(x) for x in unique_errors}
    logger.debug(f"Record {record_orm.id} has {len(error_counts)} types of errors:")
    for k, v in error_counts.items():
        logger.debug(f"    {k}: {v}")

    # Map to more general error categories
    error_counts = {error_map.get(k, "unknown_error"): v for k, v in error_counts.items()}

    # Are we beyond any of the max on any?
    for err, count in error_counts.items():
        if count > getattr(config, err, 0):
            logger.debug(f"Not auto-resetting record {record_orm.id} - has {count} errors of type {err}")
            return False

    # All good I guess
    return True



"""Singlepoint spec hashing

Revision ID: 6b758fd53ff0
Revises: 8263992eb6c8
Create Date: 2024-12-17 10:05:11.756731

"""

import sqlalchemy as sa
from alembic import op

from qcfractal.alembic.migration_helpers.hashing import hash_dict_1

# revision identifiers, used by Alembic.
revision = "6b758fd53ff0"
down_revision = "8263992eb6c8"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("qc_specification", sa.Column("specification_hash", sa.String(), nullable=True))

    # Generate hashes for existing specs
    conn = op.get_bind()
    res = conn.execute(
        sa.text(f"SELECT id, program, driver, method, basis, keywords, protocols FROM qc_specification;")
    )
    all_spec = res.fetchall()

    for spec_id, program, driver, method, basis, keywords, protocols in all_spec:
        d = {
            "program": program,
            "driver": driver,
            "method": method,
            "basis": basis,
            "keywords": keywords,
            "protocols": protocols,
        }
        h = hash_dict_1(d)
        op.execute(sa.text(f"""UPDATE qc_specification SET specification_hash = '{h}' WHERE id = {spec_id};"""))

    op.alter_column("qc_specification", "specification_hash", nullable=False)
    op.drop_constraint("ux_qc_specification_keys", "qc_specification", type_="unique")
    op.drop_column("qc_specification", "keywords_hash")
    op.create_unique_constraint("ux_qc_specification_specification_hash", "qc_specification", ["specification_hash"])
    # ### end Alembic commands ###


def downgrade():
    raise RuntimeError("Downgrade not supported.")



from __future__ import annotations

from typing import TYPE_CHECKING, Tuple, Optional, Dict, List, Union, Any

try:
    import pydantic.v1 as pydantic
except ImportError:
    import pydantic
from qcarchivetesting.helpers import read_procedure_data, read_record_data, find_test_data
from qcelemental.models import (
    Molecule,
    FailedOperation,
    ComputeError,
    AtomicResult as QCEl_AtomicResult,
    OptimizationResult as QCEl_OptimizationResult,
)

from qcfractal.components.neb.record_db_models import NEBRecordORM
from qcfractal.testing_helpers import run_service
from qcportal.generic_result import GenericTaskResult
from qcportal.neb import NEBSpecification, NEBKeywords, NEBRecord
from qcportal.record_models import PriorityEnum, RecordStatusEnum, RecordTask
from qcportal.singlepoint import SinglepointProtocols, QCSpecification
from qcportal.utils import recursive_normalizer, hash_dict

if TYPE_CHECKING:
    from qcfractal.db_socket import SQLAlchemySocket
    from qcportal.managers import ManagerName


all_test_data = find_test_data("neb_*")
all_includes = ["initial_chain", "singlepoints", "optimizations", "molecule", "comments", "initial_molecule"]

test_specs = [
    NEBSpecification(
        program="geometric",
        keywords=NEBKeywords(
            images=11,
            spring_constant=0.1,
        ),
        singlepoint_specification=QCSpecification(
            program="psi4",
            keywords={"k": "value"},
            driver="deferred",
            method="b3lyp",
            basis="6-31g",
            protocols=SinglepointProtocols(wavefunction="all"),
        ),
    ),
    NEBSpecification(
        program="geometric",
        keywords=NEBKeywords(
            images=11,
            spring_constant=0.5,
        ),
        singlepoint_specification=QCSpecification(
            program="psi4",
            keywords={"k": "value"},
            driver="deferred",
            method="CCSD(T)",
            basis="def2-tzvp",
            protocols=SinglepointProtocols(wavefunction="all"),
        ),
    ),
]


def compare_neb_specs(
    input_spec: Union[NEBSpecification, Dict[str, Any]],
    output_spec: Union[NEBSpecification, Dict[str, Any]],
) -> bool:
    if isinstance(input_spec, dict):
        input_spec = NEBSpecification(**input_spec)
    if isinstance(output_spec, dict):
        output_spec = NEBSpecification(**output_spec)

    return input_spec == output_spec


def generate_task_key(task: RecordTask):
    if task.function in ("qcengine.compute", "qcengine.compute_procedure"):
        inp_data = task.function_kwargs["input_data"]

        if inp_data["schema_name"] == "qcschema_optimization_input":
            record_type = "optimization"
            mol_hash = inp_data["initial_molecule"]["identifiers"]["molecule_hash"]
        elif inp_data["schema_name"] == "qcschema_input":
            # NEB uses both gradient and hessians on the same molecule
            record_type = "singlepoint" + "|" + inp_data["driver"]
            mol_hash = inp_data["molecule"]["identifiers"]["molecule_hash"]
        else:
            raise RuntimeError(f"Unknown result type: {inp_data['schema_name']}")

        return record_type + "|" + mol_hash
    else:
        # generic service subtask (nextchain)
        # hash function + kwargs
        normalized = recursive_normalizer(task.function_kwargs, digits=6)
        return task.function + "|" + hash_dict(normalized)


def load_procedure_data(
    name: str,
) -> Tuple[
    NEBSpecification, List[Molecule], Dict[str, Union[QCEl_AtomicResult, QCEl_OptimizationResult, GenericTaskResult]]
]:
    data = read_procedure_data(name)

    return (
        pydantic.parse_obj_as(NEBSpecification, data["specification"]),
        pydantic.parse_obj_as(List[Molecule], data["initial_chain"]),
        pydantic.parse_obj_as(
            Dict[str, Union[QCEl_AtomicResult, QCEl_OptimizationResult, GenericTaskResult]], data["results"]
        ),
    )


def load_record_data(name: str) -> NEBRecord:
    data = read_record_data(name)
    return NEBRecord(**data)


def submit_procedure_data(
    storage_socket: SQLAlchemySocket,
    name: str,
    compute_tag: Optional[str] = "*",
    compute_priority: PriorityEnum = PriorityEnum.normal,
) -> Tuple[int, Dict[str, Any]]:
    input_spec, initial_chain, result = load_procedure_data(name)
    meta, record_ids = storage_socket.records.neb.add(
        [initial_chain], input_spec, compute_tag, compute_priority, None, True
    )
    assert meta.success
    assert len(record_ids) == 1
    assert meta.n_inserted == 1

    return record_ids[0], result


def run_procedure_data(
    storage_socket: SQLAlchemySocket,
    manager_name: ManagerName,
    name: str,
    compute_tag: Optional[str] = "*",
    compute_priority: PriorityEnum = PriorityEnum.normal,
    end_status: RecordStatusEnum = RecordStatusEnum.complete,
):
    record_id, result = submit_procedure_data(storage_socket, name, compute_tag, compute_priority)

    with storage_socket.session_scope() as session:
        record = session.get(NEBRecordORM, record_id)
        assert record.status == RecordStatusEnum.waiting

    if end_status == RecordStatusEnum.error:
        failed_op = FailedOperation(
            error=ComputeError(error_type="test_error", error_message="this is just a test error"),
        )
        result = {x: failed_op for x in result}

    finished, n_optimizations = run_service(storage_socket, manager_name, record_id, generate_task_key, result, 200)
    assert finished

    with storage_socket.session_scope() as session:
        record = session.get(NEBRecordORM, record_id)
        assert record.status == end_status

    return record_id



"""
Tests the tasks socket (claiming & returning data)
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from qcarchivetesting.testing_classes import QCATestingSnowflake
from qcfractal.components.managers.db_models import ComputeManagerORM
from qcfractal.components.optimization.testing_helpers import load_procedure_data as load_opt_procedure_data
from qcfractal.components.record_db_models import BaseRecordORM
from qcfractal.components.singlepoint.testing_helpers import load_procedure_data as load_sp_procedure_data
from qcfractal.components.torsiondrive.testing_helpers import (
    load_procedure_data as load_td_procedure_data,
    generate_task_key as generate_td_task_key,
)
from qcfractal.testing_helpers import run_service
from qcfractalcompute.compress import compress_result
from qcportal.exceptions import ComputeManagerError
from qcportal.managers import ManagerName
from qcportal.record_models import PriorityEnum
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from qcfractal.db_socket import SQLAlchemySocket
    from sqlalchemy.orm.session import Session

input_spec_1, molecule_1, result_data_1 = load_sp_procedure_data("sp_psi4_water_energy")
input_spec_2, molecule_2, result_data_2 = load_sp_procedure_data("sp_psi4_water_gradient")
input_spec_3, molecule_3, result_data_3 = load_sp_procedure_data("sp_psi4_water_hessian")
input_spec_4, molecule_4, result_data_4 = load_opt_procedure_data("opt_psi4_benzene")
input_spec_5, molecule_5, result_data_5 = load_sp_procedure_data("sp_psi4_benzene_energy_1")
input_spec_6, molecule_6, result_data_6 = load_sp_procedure_data("sp_psi4_benzene_energy_2")
input_spec_7, molecule_7, result_data_7 = load_sp_procedure_data("sp_rdkit_benzene_energy")


def test_task_socket_update_manager_time(storage_socket: SQLAlchemySocket, session: Session):
    # Test that claiming & returning updates the modified_on time of the manager
    mname1 = ManagerName(cluster="test_cluster", hostname="a_host1", uuid="1234-5678-1234-5678")

    mprog1 = {"qcengine": ["unknown"], "psi4": ["unknown"], "geometric": ["v3.0"]}

    mid_1 = storage_socket.managers.activate(
        name_data=mname1,
        manager_version="v2.0",
        username="bill",
        programs=mprog1,
        compute_tags=["tag1"],
    )

    meta, id_1 = storage_socket.records.singlepoint.add(
        [molecule_1], input_spec_1, "tag1", PriorityEnum.low, None, True
    )

    # claim up to two tasks
    t0 = now_at_utc()
    tasks = storage_socket.tasks.claim_tasks(mname1.fullname, mprog1, ["tag1"], 2)
    t1 = now_at_utc()
    assert len(tasks) == 1

    # Check assignments
    session.expire_all()

    m1 = session.get(ComputeManagerORM, mid_1)
    assert m1.claimed == 1
    assert t0 < m1.modified_on < t1

    # Claiming but getting nothing also updates the time
    t0 = now_at_utc()
    tasks0 = storage_socket.tasks.claim_tasks(mname1.fullname, mprog1, ["tag1"], 2)
    t1 = now_at_utc()
    assert len(tasks0) == 0

    session.expire_all()
    m1 = session.get(ComputeManagerORM, mid_1)
    assert m1.claimed == 1
    assert t0 < m1.modified_on < t1

    # Return the task data
    t0 = now_at_utc()
    rmeta = storage_socket.tasks.update_finished(
        mname1.fullname,
        {
            tasks[0]["id"]: compress_result(result_data_1.dict()),
        },
    )
    t1 = now_at_utc()

    assert rmeta.n_accepted == 1
    session.expire_all()
    m1 = session.get(ComputeManagerORM, mid_1)
    assert t0 < m1.modified_on < t1


def test_task_socket_claim_mixed(storage_socket: SQLAlchemySocket, session: Session):
    mname1 = ManagerName(cluster="test_cluster", hostname="a_host1", uuid="1234-5678-1234-5678")
    mname2 = ManagerName(cluster="test_cluster", hostname="a_host2", uuid="2234-5678-1234-5678")
    mname3 = ManagerName(cluster="test_cluster", hostname="a_host3", uuid="3234-5678-1234-5678")
    mname4 = ManagerName(cluster="test_cluster", hostname="a_host4", uuid="4234-5678-1234-5678")

    mprog1 = {"qcengine": ["unknown"], "psi4": ["unknown"], "geometric": ["v3.0"]}
    mprog2 = {"qcengine": ["unknown"], "psi4": ["unknown"], "geometric": ["v3.0"]}
    mprog3 = {"qcengine": ["unknown"], "psi4": ["unknown"], "geometric": ["v3.0"]}
    mprog4 = {"qcengine": ["unknown"], "psi4": ["unknown"], "geometric": ["v3.0"], "rdkit": ["v1.0"]}

    mid_1 = storage_socket.managers.activate(
        name_data=mname1,
        manager_version="v2.0",
        username="bill",
        programs=mprog1,
        compute_tags=["tag1"],
    )
    mid_2 = storage_socket.managers.activate(
        name_data=mname2,
        manager_version="v2.0",
        username="bill",
        programs=mprog2,
        compute_tags=["*"],
    )
    mid_3 = storage_socket.managers.activate(
        name_data=mname3,
        manager_version="v2.0",
        username="bill",
        programs=mprog3,
        compute_tags=["tag3", "tag1"],
    )
    mid_4 = storage_socket.managers.activate(
        name_data=mname4,
        manager_version="v2.0",
        username="bill",
        programs=mprog4,
        compute_tags=["tag3", "*"],
    )

    meta, id_1 = storage_socket.records.singlepoint.add(
        [molecule_1], input_spec_1, "tag1", PriorityEnum.low, None, True
    )
    meta, id_2 = storage_socket.records.singlepoint.add(
        [molecule_2], input_spec_2, "tag2", PriorityEnum.normal, None, True
    )
    meta, id_3 = storage_socket.records.singlepoint.add(
        [molecule_3], input_spec_3, "tag1", PriorityEnum.high, None, True
    )
    meta, id_4 = storage_socket.records.optimization.add(
        [molecule_4], input_spec_4, "tag3", PriorityEnum.normal, None, True
    )
    meta, id_5 = storage_socket.records.singlepoint.add(
        [molecule_5], input_spec_5, "tag1", PriorityEnum.normal, None, True
    )
    meta, id_6 = storage_socket.records.singlepoint.add(
        [molecule_6], input_spec_6, "tag6", PriorityEnum.high, None, True
    )
    meta, id_7 = storage_socket.records.singlepoint.add(
        [molecule_7], input_spec_7, "tag1", PriorityEnum.high, None, True
    )

    all_id = id_1 + id_2 + id_3 + id_4 + id_5 + id_6 + id_7
    recs = []
    for rid in all_id:
        rec = session.get(BaseRecordORM, rid)
        recs.append(rec)

    # claim up to two tasks
    # should find the high and normal priority one, but not the one
    # requiring rdkit
    tasks = storage_socket.tasks.claim_tasks(mname1.fullname, mprog1, ["tag1"], 2)
    assert len(tasks) == 2
    assert tasks[0]["id"] == recs[2].task.id
    assert tasks[1]["id"] == recs[4].task.id

    # manager 4 should find tag3, and then #6 (highest priority left)
    tasks = storage_socket.tasks.claim_tasks(mname4.fullname, mprog4, ["tag3", "*"], 2)
    assert len(tasks) == 2
    assert tasks[0]["id"] == recs[3].task.id
    assert tasks[1]["id"] == recs[5].task.id

    # manager3 should find the only remaining tag1 that isn't rdkit
    tasks = storage_socket.tasks.claim_tasks(mname3.fullname, mprog3, ["tag3", "tag1"])
    assert len(tasks) == 1
    assert tasks[0]["id"] == recs[0].task.id

    # manager 2 only finds #2 - doesn't have rdkit
    tasks = storage_socket.tasks.claim_tasks(mname2.fullname, mprog2, ["*"], 20)
    assert len(tasks) == 1
    assert tasks[0]["id"] == recs[1].task.id

    # manager 4 can finally get the last one
    tasks = storage_socket.tasks.claim_tasks(mname4.fullname, mprog4, ["tag3", "*"], 20)
    assert len(tasks) == 1
    assert tasks[0]["id"] == recs[6].task.id

    # Check assignments
    session.expire_all()
    recs = []
    for rid in all_id:
        rec = session.get(BaseRecordORM, rid)
        recs.append(rec)

    assert recs[0].manager_name == mname3.fullname
    assert recs[1].manager_name == mname2.fullname
    assert recs[2].manager_name == mname1.fullname
    assert recs[3].manager_name == mname4.fullname
    assert recs[4].manager_name == mname1.fullname
    assert recs[5].manager_name == mname4.fullname
    assert recs[6].manager_name == mname4.fullname

    assert session.get(ComputeManagerORM, mid_1).claimed == 2
    assert session.get(ComputeManagerORM, mid_2).claimed == 1
    assert session.get(ComputeManagerORM, mid_3).claimed == 1
    assert session.get(ComputeManagerORM, mid_4).claimed == 3


def test_task_socket_claim_priority(storage_socket: SQLAlchemySocket, session: Session):
    mname1 = ManagerName(cluster="test_cluster", hostname="a_host1", uuid="1234-5678-1234-5678")
    mprog1 = {"qcengine": ["unknown"], "psi4": ["unknown"], "geometric": ["v3.0"]}
    storage_socket.managers.activate(
        name_data=mname1,
        manager_version="v2.0",
        username="bill",
        programs=mprog1,
        compute_tags=["tag1"],
    )

    meta, id_1 = storage_socket.records.singlepoint.add(
        [molecule_1], input_spec_1, "tag1", PriorityEnum.low, None, True
    )
    meta, id_2 = storage_socket.records.singlepoint.add(
        [molecule_2], input_spec_2, "tag1", PriorityEnum.normal, None, True
    )
    meta, id_3 = storage_socket.records.singlepoint.add(
        [molecule_3], input_spec_3, "tag1", PriorityEnum.high, None, True
    )
    meta, id_4 = storage_socket.records.optimization.add(
        [molecule_4],
        input_spec_4,
        "tag1",
        PriorityEnum.normal,
        None,
        True,
    )
    meta, id_5 = storage_socket.records.singlepoint.add(
        [molecule_5], input_spec_5, "tag1", PriorityEnum.normal, None, True
    )
    meta, id_6 = storage_socket.records.singlepoint.add(
        [molecule_6], input_spec_6, "tag1", PriorityEnum.high, None, True
    )

    all_id = id_1 + id_2 + id_3 + id_4 + id_5 + id_6
    recs = []
    for rid in all_id:
        rec = session.get(BaseRecordORM, rid)
        recs.append(rec)

    # highest priority should be first, then by modified date
    tasks = storage_socket.tasks.claim_tasks(mname1.fullname, mprog1, ["tag1"], 3)
    assert len(tasks) == 3
    assert tasks[0]["id"] == recs[2].task.id
    assert tasks[1]["id"] == recs[5].task.id
    assert tasks[2]["id"] == recs[1].task.id

    # Now normal then low
    tasks = storage_socket.tasks.claim_tasks(mname1.fullname, mprog1, ["tag1"], 3)
    assert len(tasks) == 3
    assert tasks[0]["id"] == recs[3].task.id
    assert tasks[1]["id"] == recs[4].task.id
    assert tasks[2]["id"] == recs[0].task.id


def test_task_socket_claim_tag(storage_socket: SQLAlchemySocket, session: Session):
    mname1 = ManagerName(cluster="test_cluster", hostname="a_host1", uuid="1234-5678-1234-5678")
    mprog1 = {"qcengine": ["unknown"], "psi4": ["unknown"], "geometric": ["v3.0"]}
    storage_socket.managers.activate(
        name_data=mname1,
        manager_version="v2.0",
        username="bill",
        programs=mprog1,
        compute_tags=["tag3", "tag1"],
    )

    meta, id_1 = storage_socket.records.singlepoint.add(
        [molecule_1], input_spec_1, "TAg1", PriorityEnum.normal, None, True
    )
    meta, id_2 = storage_socket.records.singlepoint.add(
        [molecule_2], input_spec_2, "tag2", PriorityEnum.normal, None, True
    )
    meta, id_3 = storage_socket.records.singlepoint.add(
        [molecule_3], input_spec_3, "*", PriorityEnum.normal, None, True
    )
    meta, id_4 = storage_socket.records.optimization.add(
        [molecule_4], input_spec_4, "tag3", PriorityEnum.normal, None, True
    )
    meta, id_5 = storage_socket.records.singlepoint.add(
        [molecule_5], input_spec_5, "tag1", PriorityEnum.normal, None, True
    )

    all_id = id_1 + id_2 + id_3 + id_4 + id_5
    recs = []
    for rid in all_id:
        rec = session.get(BaseRecordORM, rid)
        recs.append(rec)

    # tag3 should be first, then tag1
    # Change capitalization to test case insensitivity
    tasks = storage_socket.tasks.claim_tasks(mname1.fullname, mprog1, ["Tag3", "Tag1"], 2)
    assert len(tasks) == 2
    assert tasks[0]["id"] == recs[3].task.id
    assert tasks[1]["id"] == recs[0].task.id

    tasks = storage_socket.tasks.claim_tasks(mname1.fullname, mprog1, ["tag3", "tag1"], 3)
    assert len(tasks) == 1
    assert tasks[0]["id"] == recs[4].task.id


def test_task_socket_claim_tag_wildcard(storage_socket: SQLAlchemySocket, session: Session):
    mname1 = ManagerName(cluster="test_cluster", hostname="a_host1", uuid="1234-5678-1234-5678")
    mprog1 = {"qcengine": ["unknown"], "psi4": ["unknown"], "geometric": ["v3.0"]}
    storage_socket.managers.activate(
        name_data=mname1,
        manager_version="v2.0",
        username="bill",
        programs=mprog1,
        compute_tags=["TAG3", "*"],
    )

    meta, id_1 = storage_socket.records.singlepoint.add(
        [molecule_1], input_spec_1, "tag1", PriorityEnum.normal, None, True
    )
    meta, id_2 = storage_socket.records.singlepoint.add(
        [molecule_2], input_spec_2, "tag2", PriorityEnum.normal, None, True
    )
    meta, id_3 = storage_socket.records.singlepoint.add(
        [molecule_3], input_spec_3, "*", PriorityEnum.normal, None, True
    )
    meta, id_4 = storage_socket.records.optimization.add(
        [molecule_4], input_spec_4, "taG3", PriorityEnum.normal, None, True
    )
    meta, id_5 = storage_socket.records.singlepoint.add(
        [molecule_5], input_spec_5, "tag1", PriorityEnum.normal, None, True
    )

    all_id = id_1 + id_2 + id_3 + id_4 + id_5
    recs = []
    for rid in all_id:
        rec = session.get(BaseRecordORM, rid)
        recs.append(rec)

    # tag3 should be first, then any task (in order)
    tasks = storage_socket.tasks.claim_tasks(mname1.fullname, mprog1, ["tag3", "*"], 2)
    assert len(tasks) == 2
    assert tasks[0]["id"] == recs[3].task.id
    assert tasks[1]["id"] == recs[0].task.id

    tasks = storage_socket.tasks.claim_tasks(mname1.fullname, mprog1, ["tag3", "*"], 3)
    assert len(tasks) == 3
    assert tasks[0]["id"] == recs[1].task.id
    assert tasks[1]["id"] == recs[2].task.id
    assert tasks[2]["id"] == recs[4].task.id


def test_task_socket_claim_tag_wildcard_strict(postgres_server, pytestconfig):

    pg_harness = postgres_server.get_new_harness("claim_tag_wildcard_strict")
    encoding = pytestconfig.getoption("--client-encoding")
    with QCATestingSnowflake(pg_harness, encoding=encoding, extra_config={"strict_compute_tags": True}) as snowflake:
        storage_socket = snowflake.get_storage_socket()

        mname1 = ManagerName(cluster="test_cluster", hostname="a_host1", uuid="1234-5678-1234-5678")
        mprog1 = {"qcengine": ["unknown"], "psi4": ["unknown"], "geometric": ["v3.0"]}
        storage_socket.managers.activate(
            name_data=mname1,
            manager_version="v2.0",
            username="bill",
            programs=mprog1,
            compute_tags=["TAG3", "*"],
        )

        meta, id_1 = storage_socket.records.singlepoint.add(
            [molecule_1], input_spec_1, "tag1", PriorityEnum.normal, None, True
        )
        meta, id_2 = storage_socket.records.singlepoint.add(
            [molecule_2], input_spec_2, "tag2", PriorityEnum.normal, None, True
        )
        meta, id_3 = storage_socket.records.singlepoint.add(
            [molecule_3], input_spec_3, "*", PriorityEnum.normal, None, True
        )
        meta, id_4 = storage_socket.records.optimization.add(
            [molecule_4], input_spec_4, "taG3", PriorityEnum.normal, None, True
        )
        meta, id_5 = storage_socket.records.singlepoint.add(
            [molecule_5], input_spec_5, "tag1", PriorityEnum.normal, None, True
        )

        all_id = id_1 + id_2 + id_3 + id_4 + id_5

        client = snowflake.client()
        recs = client.get_records(all_id, include=["task"])

        # tag3 should be first, then only the * task (in order)
        tasks = storage_socket.tasks.claim_tasks(mname1.fullname, mprog1, ["tag3", "*"], 2)
        assert len(tasks) == 2
        assert tasks[0]["id"] == recs[3].task.id
        assert tasks[1]["id"] == recs[2].task.id

        tasks = storage_socket.tasks.claim_tasks(mname1.fullname, mprog1, ["tag3", "*"], 3)
        assert len(tasks) == 0


def test_task_socket_claim_program(storage_socket: SQLAlchemySocket, session: Session):
    mname1 = ManagerName(cluster="test_cluster", hostname="a_host1", uuid="1234-5678-1234-5678")

    # Managers should not send programs with uppercase characters, but be forgiving
    mprog1 = {"QCengine": ["unknown"], "psi4": ["unknown"], "Geometric": ["V3.0"]}
    storage_socket.managers.activate(
        name_data=mname1,
        manager_version="v2.0",
        username="bill",
        programs=mprog1,
        compute_tags=["*"],
    )

    meta, id_7 = storage_socket.records.singlepoint.add(
        [molecule_7], input_spec_7, "tag1", PriorityEnum.normal, None, True
    )
    meta, id_1 = storage_socket.records.singlepoint.add(
        [molecule_1], input_spec_1, "tag1", PriorityEnum.normal, None, True
    )
    meta, id_2 = storage_socket.records.singlepoint.add(
        [molecule_2], input_spec_2, "tag1", PriorityEnum.normal, None, True
    )

    all_id = id_7 + id_1 + id_2
    recs = []
    for rid in all_id:
        rec = session.get(BaseRecordORM, rid)
        recs.append(rec)

    # claim all tasks. But it shouldn't claim #7
    tasks = storage_socket.tasks.claim_tasks(mname1.fullname, mprog1, ["*"], 100)
    assert len(tasks) == 2
    assert tasks[0]["id"] == recs[1].task.id
    assert tasks[1]["id"] == recs[2].task.id


def test_task_socket_claim_program_subset(storage_socket: SQLAlchemySocket, session: Session):
    mname1 = ManagerName(cluster="test_cluster", hostname="a_host1", uuid="1234-5678-1234-5678")
    mprog1 = {"qcengine": ["unknown"], "psi4": ["unknown"], "geometric": ["v3.0"], "rdkit": ["unknown"]}
    storage_socket.managers.activate(
        name_data=mname1,
        manager_version="v2.0",
        username="bill",
        programs=mprog1,
        compute_tags=["*"],
    )

    meta, id_7 = storage_socket.records.singlepoint.add(
        [molecule_7], input_spec_7, "tag1", PriorityEnum.normal, None, True
    )
    meta, id_1 = storage_socket.records.singlepoint.add(
        [molecule_1], input_spec_1, "tag1", PriorityEnum.normal, None, True
    )
    meta, id_2 = storage_socket.records.singlepoint.add(
        [molecule_2], input_spec_2, "tag1", PriorityEnum.normal, None, True
    )

    all_id = id_7 + id_1 + id_2
    recs = []
    for rid in all_id:
        rec = session.get(BaseRecordORM, rid)
        recs.append(rec)

    # claim all tasks. Shouldn't claim #7 (rdkit not included)
    claim_prog = {"qcengine": ["unknown"], "psi4": ["unknown"], "geometric": ["v3.0"]}
    tasks = storage_socket.tasks.claim_tasks(mname1.fullname, claim_prog, ["*"], 100)
    assert len(tasks) == 2
    assert tasks[0]["id"] == recs[1].task.id
    assert tasks[1]["id"] == recs[2].task.id

    # now claim the rdkit task
    claim_prog = {"qcengine": ["unknown"], "rdkit": ["unknown"]}
    tasks = storage_socket.tasks.claim_tasks(mname1.fullname, claim_prog, ["*"], 100)
    assert len(tasks) == 1
    assert tasks[0]["id"] == recs[0].task.id


def test_task_socket_claim_program_missing(storage_socket: SQLAlchemySocket):
    mname1 = ManagerName(cluster="test_cluster", hostname="a_host1", uuid="1234-5678-1234-5678")
    mprog1 = {"qcengine": ["unknown"], "psi4": ["unknown"], "geometric": ["v3.0"], "rdkit": ["unknown"]}
    storage_socket.managers.activate(
        name_data=mname1,
        manager_version="v2.0",
        username="bill",
        programs=mprog1,
        compute_tags=["*"],
    )

    claim_prog = {}

    with pytest.raises(ComputeManagerError, match="did not send any valid programs to claim"):
        storage_socket.tasks.claim_tasks(mname1.fullname, claim_prog, ["*"], 100)


def test_task_socket_claim_tags_missing(storage_socket: SQLAlchemySocket):
    mname1 = ManagerName(cluster="test_cluster", hostname="a_host1", uuid="1234-5678-1234-5678")
    mprog1 = {"qcengine": ["unknown"], "psi4": ["unknown"], "geometric": ["v3.0"], "rdkit": ["unknown"]}
    storage_socket.managers.activate(
        name_data=mname1,
        manager_version="v2.0",
        username="bill",
        programs=mprog1,
        compute_tags=["*"],
    )

    claim_prog = {"qcengine": ["unknown"], "rdkit": ["unknown"]}

    with pytest.raises(ComputeManagerError, match="did not send any valid queue tags to claim"):
        storage_socket.tasks.claim_tasks(mname1.fullname, claim_prog, [], 100)


def test_task_socket_claim_service_order(storage_socket: SQLAlchemySocket, session: Session):
    input_spec_1, molecules_1, result_data_1 = load_td_procedure_data("td_H2O2_psi4_pbe")
    input_spec_2, molecules_2, result_data_2 = load_td_procedure_data("td_H2O2_mopac_pm6")

    mname1 = ManagerName(cluster="test_cluster", hostname="a_host1", uuid="1234-5678-1234-5678")

    mprog1 = {"qcengine": ["unknown"], "psi4": ["unknown"], "geometric": ["v3.0"], "mopac": ["unknown"]}

    mid_1 = storage_socket.managers.activate(
        name_data=mname1,
        manager_version="v2.0",
        username="bill",
        programs=mprog1,
        compute_tags=["*"],
    )

    meta_1, id_1 = storage_socket.records.torsiondrive.add(
        [molecules_1], input_spec_1, True, "tag1", PriorityEnum.low, None, True
    )
    meta_2, id_2 = storage_socket.records.torsiondrive.add(
        [molecules_2], input_spec_2, True, "tag1", PriorityEnum.low, None, True
    )

    id_1 = id_1[0]
    id_2 = id_2[0]

    finished, n_records = run_service(storage_socket, mname1, id_1, generate_td_task_key, result_data_1, 1)
    finished, n_records = run_service(storage_socket, mname1, id_2, generate_td_task_key, result_data_2, 1)
    assert finished is False

    # Iterate the second service first
    with storage_socket.session_scope() as session:
        service_orm = session.get(BaseRecordORM, id_2).service
        finished = storage_socket.records.iterate_service(session, service_orm)
        assert finished is False

    with storage_socket.session_scope() as session:
        service_orm = session.get(BaseRecordORM, id_1).service
        finished = storage_socket.records.iterate_service(session, service_orm)
        assert finished is False

    # The manager should claim tasks belonging to that first service, not the second
    tasks = storage_socket.tasks.claim_tasks(mname1.fullname, mprog1, ["*"], 1)
    assert len(tasks) == 1

    with storage_socket.session_scope() as session:
        service_orm = session.get(BaseRecordORM, id_1).service
        dep_ids = [d.record_id for d in service_orm.dependencies]
        assert tasks[0]["record_id"] in dep_ids

        service_orm_2 = session.get(BaseRecordORM, id_2).service
        dep_ids_2 = [d.record_id for d in service_orm_2.dependencies]
        assert tasks[0]["record_id"] not in dep_ids_2



"""remove project molecules

Revision ID: a6a708376a1e
Revises: 4d5ad63ad478
Create Date: 2025-04-27 10:42:26.756827

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "a6a708376a1e"
down_revision = "4d5ad63ad478"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_project_molecule_molecule_id", table_name="project_molecule")
    op.drop_table("project_molecule")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "project_molecule",
        sa.Column("project_id", sa.INTEGER(), autoincrement=False, nullable=False),
        sa.Column("molecule_id", sa.INTEGER(), autoincrement=False, nullable=False),
        sa.Column("name", sa.VARCHAR(), autoincrement=False, nullable=False),
        sa.Column(
            "lname",
            sa.VARCHAR(length=100),
            sa.Computed("lower((name)::text)", persisted=True),
            autoincrement=False,
            nullable=False,
        ),
        sa.Column("description", sa.VARCHAR(), autoincrement=False, nullable=False),
        sa.Column("tags", postgresql.JSON(astext_type=sa.Text()), autoincrement=False, nullable=False),
        sa.ForeignKeyConstraint(["molecule_id"], ["molecule.id"], name="project_molecule_molecule_id_fkey"),
        sa.ForeignKeyConstraint(
            ["project_id"], ["project.id"], name="project_molecule_project_id_fkey", ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("project_id", "lname", name="project_molecule_pkey"),
    )
    op.create_index("ix_project_molecule_molecule_id", "project_molecule", ["molecule_id"], unique=False)
    # ### end Alembic commands ###



from __future__ import annotations

import ipaddress

import pytest

from qcarchivetesting import load_ip_test_data, ip_tests_enabled
from qcarchivetesting.testing_classes import QCATestingSnowflake
from qcportal.serverinfo.models import AccessLogQueryFilters
from qcportal.utils import now_at_utc

# First part of the tuple is the ip address
# second is the range, as stored in the MaxMind test JSON file
test_ips = [
    ("81.2.69.142", "::81.2.69.192/124"),
    ("175.16.199.2", "::175.16.199.0/120"),
    ("89.160.20.113", "::89.160.20.112/124"),
    ("2001:02b8:0000:0000:0000:0000:0000:123f", "2001:2b8::/32"),
    ("10.0.0.1", None),
    ("2.125.160.217", "::2.125.160.216/125"),
]


@pytest.mark.skipif(not ip_tests_enabled, reason="Test GeoIP data not found")
def test_serverinfo_socket_save_access(secure_snowflake: QCATestingSnowflake):
    storage_socket = secure_snowflake.get_storage_socket()

    time_0 = now_at_utc()

    ip_data = load_ip_test_data()

    admin_id = storage_socket.users.get("admin_user")["id"]
    read_id = storage_socket.users.get("read_user")["id"]
    monitor_id = storage_socket.users.get("monitor_user")["id"]

    userid_map = {admin_id: "admin_user", read_id: "read_user", monitor_id: "monitor_user"}

    access1 = {
        "module": "api",
        "method": "GET",
        "full_uri": "/api/v1/datasets",
        "ip_address": test_ips[0][0],
        "user_agent": "Fake user agent",
        "request_duration": 0.24,
        "user_id": admin_id,
        "request_bytes": 123,
        "response_bytes": 18273,
    }

    access2 = {
        "module": "api",
        "method": "POST",
        "full_uri": "/api/v1/records",
        "ip_address": test_ips[1][0],
        "user_agent": "Fake user agent",
        "request_duration": 0.45,
        "user_id": read_id,
        "request_bytes": 456,
        "response_bytes": 12671,
    }

    access3 = {
        "module": "api",
        "method": "GET",
        "full_uri": "/api/v1/me",
        "ip_address": test_ips[2][0],
        "user_agent": "Fake user agent",
        "request_duration": 0.01,
        "user_id": read_id,
        "request_bytes": 789,
        "response_bytes": 1975,
    }

    access4 = {
        "module": "api",
        "method": "PUT",
        "full_uri": "/api/v1/users",
        "ip_address": test_ips[3][0],
        "user_agent": "Fake user agent",
        "request_duration": 2.18,
        "user_id": admin_id,
        "request_bytes": 101112,
        "response_bytes": 10029,
    }

    access5 = {
        "module": "compute",
        "full_uri": "/compute/v1/activate",
        "method": "POST",
        "ip_address": test_ips[4][0],
        "user_agent": "Fake user agent",
        "request_duration": 3.12,
        "user_id": monitor_id,
        "request_bytes": 151617,
        "response_bytes": 2719,
    }

    access6 = {
        "module": "auth",
        "full_uri": "/auth/v1/login",
        "method": "POST",
        "ip_address": test_ips[5][0],
        "user_agent": "Fake user agent",
        "request_duration": 1.28,
        "user_id": read_id,
        "request_bytes": 131415,
        "response_bytes": 718723,
    }

    all_accesses = [access1, access2, access3, access4, access5, access6]
    storage_socket.serverinfo.save_access(access1)
    storage_socket.serverinfo.save_access(access2)
    storage_socket.serverinfo.save_access(access3)
    storage_socket.serverinfo.save_access(access4)
    storage_socket.serverinfo.save_access(access5)
    storage_socket.serverinfo.save_access(access6)

    time_1 = now_at_utc()

    # Update the IP addresses with geo data
    with storage_socket.session_scope() as session:
        storage_socket.serverinfo.geolocate_accesses(session)

    accesses = storage_socket.serverinfo.query_access_log(AccessLogQueryFilters())
    assert len(accesses) >= 6

    accesses = storage_socket.serverinfo.query_access_log(AccessLogQueryFilters(before=time_1, after=time_0))
    assert len(accesses) == 6

    # Order should be latest access first
    assert accesses[0]["timestamp"] > accesses[1]["timestamp"]
    assert accesses[1]["timestamp"] > accesses[2]["timestamp"]
    assert accesses[2]["timestamp"] > accesses[3]["timestamp"]
    assert accesses[3]["timestamp"] > accesses[4]["timestamp"]
    assert accesses[4]["timestamp"] > accesses[5]["timestamp"]

    # These are ordered descending (newest accesses first). Reverse the order for testing
    for i, (ac_in, ac_db) in enumerate(zip(all_accesses, reversed(accesses))):
        assert ac_in["module"] == ac_db["module"]
        assert ac_in["method"] == ac_db["method"]
        assert ac_in["full_uri"] == ac_db["full_uri"]

        # IPV6 can have different representations, so use the ipaddress library for comparison
        assert ipaddress.ip_address(ac_in["ip_address"]) == ipaddress.ip_address(ac_db["ip_address"])

        assert ac_in["user_agent"] == ac_db["user_agent"]
        assert ac_in["request_duration"] == ac_db["request_duration"]
        assert userid_map[ac_in["user_id"]] == ac_db["user"]
        assert ac_in["request_bytes"] == ac_db["request_bytes"]
        assert ac_in["response_bytes"] == ac_db["response_bytes"]

        ip_lookup_key = test_ips[i][1]
        ip_ref_data = ip_data.get(ip_lookup_key)

        if ip_ref_data is not None:
            assert ac_db["country_code"] == ip_ref_data["country"]["iso_code"]
            assert ac_db["ip_lat"] == ip_ref_data["location"]["latitude"]
            assert ac_db["ip_long"] == ip_ref_data["location"]["longitude"]

            if ac_db.get("subdivision") is not None:
                assert ac_db["subdivision"] == ip_ref_data["subdivisions"][-1]["names"]["en"]
            if ac_db.get("city") is not None:
                assert ac_db["city"] == ip_ref_data["city"]["names"]["en"]
            if ac_db.get("ip_lat") is not None:
                assert ac_db["ip_lat"] == ip_ref_data["location"]["latitude"]
            if ac_db.get("ip_long") is not None:
                assert ac_db["ip_long"] == ip_ref_data["location"]["longitude"]



from flask import current_app

from qcfractal.flask_app import storage_socket
from qcfractal.flask_app.api_v1.blueprint import api_v1
from qcfractal.flask_app.helpers import get_public_server_information
from qcfractal.flask_app.wrap_route import wrap_global_route
from qcportal.serverinfo import (
    AccessLogSummaryFilters,
    AccessLogQueryFilters,
    ErrorLogQueryFilters,
    DeleteBeforeDateBody,
)
from qcportal.utils import calculate_limit


@api_v1.route("/information", methods=["GET"])
@wrap_global_route("information", "read")
def get_information():
    return get_public_server_information()


@api_v1.route("/motd", methods=["GET"])
@wrap_global_route("information", "read")
def get_motd():
    return storage_socket.serverinfo.get_motd()


@api_v1.route("/motd", methods=["PUT"])
@wrap_global_route("information", "modify")
def set_motd(body_data: str):
    return storage_socket.serverinfo.set_motd(body_data)


@api_v1.route("/access_logs/query", methods=["POST"])
@wrap_global_route("access_log", "read")
def query_access_log_v1(body_data: AccessLogQueryFilters):
    max_limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.get_access_logs
    body_data.limit = calculate_limit(max_limit, body_data.limit)
    return storage_socket.serverinfo.query_access_log(body_data)


@api_v1.route("/access_logs/bulkDelete", methods=["POST"])
@wrap_global_route("access_log", "delete")
def delete_access_log_v1(body_data: DeleteBeforeDateBody):
    return storage_socket.serverinfo.delete_access_logs(before=body_data.before)


@api_v1.route("/access_logs/summary", methods=["GET"])
@wrap_global_route("access_log", "read")
def query_access_summary_v1(url_params: AccessLogSummaryFilters):
    return storage_socket.serverinfo.query_access_summary(url_params)


@api_v1.route("/server_errors/query", methods=["POST"])
@wrap_global_route("server_errors", "read")
def query_error_log_v1(body_data: ErrorLogQueryFilters):
    max_limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.get_error_logs
    body_data.limit = calculate_limit(max_limit, body_data.limit)
    return storage_socket.serverinfo.query_error_log(body_data)


@api_v1.route("/server_errors/bulkDelete", methods=["POST"])
@wrap_global_route("server_errors", "delete")
def delete_server_error_log_v1(body_data: DeleteBeforeDateBody):
    return storage_socket.serverinfo.delete_error_logs(before=body_data.before)



from __future__ import annotations

from sqlalchemy import Column, Integer, ForeignKey, String, ForeignKeyConstraint, Index, select
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import relationship

from qcfractal.components.dataset_db_models import BaseDatasetORM
from qcfractal.components.molecules.db_models import MoleculeORM
from qcfractal.components.neb.record_db_models import NEBRecordORM, NEBSpecificationORM
from qcfractal.db_socket import BaseORM


class NEBDatasetInitialMoleculeORM(BaseORM):
    """
    Association table neb -> initial chain
    """

    __tablename__ = "neb_dataset_molecule"

    dataset_id = Column(Integer, primary_key=True)
    entry_name = Column(String, primary_key=True)
    molecule_id = Column(Integer, ForeignKey(MoleculeORM.id), primary_key=True)
    position = Column(Integer, primary_key=True)

    # molecule = relationship(MoleculeORM)

    __table_args__ = (
        Index("ix_neb_dataset_molecule_dataset_id", "dataset_id"),
        Index("ix_neb_dataset_molecule_entry_name", "entry_name"),
        Index("ix_neb_dataset_molecule_molecule_id", "molecule_id"),
        ForeignKeyConstraint(
            ["dataset_id", "entry_name"],
            ["neb_dataset_entry.dataset_id", "neb_dataset_entry.name"],
            ondelete="cascade",
            onupdate="cascade",
        ),
    )

    _qcportal_model_excludes = ["dataset_id", "molecule_id", "entry_name", "position"]


class NEBDatasetEntryORM(BaseORM):
    __tablename__ = "neb_dataset_entry"

    dataset_id = Column(Integer, ForeignKey("neb_dataset.id", ondelete="cascade"), primary_key=True)

    name = Column(String, primary_key=True)
    comment = Column(String)

    additional_keywords = Column(JSONB, nullable=False)
    additional_singlepoint_keywords = Column(JSONB, nullable=False)
    attributes = Column(JSONB, nullable=False)

    # initial_molecule_ids = column_property(
    #   select(array_agg(NEBDatasetInitialMoleculeORM.molecule_id))
    #   .where(NEBDatasetInitialMoleculeORM.dataset_id == dataset_id)
    #   .where(NEBDatasetInitialMoleculeORM.entry_name == name)
    #   .scalar_subquery()
    # )

    initial_chain = relationship(
        MoleculeORM,
        secondary=NEBDatasetInitialMoleculeORM.__table__,
        order_by=NEBDatasetInitialMoleculeORM.__table__.c.position,
        viewonly=True,
        lazy="selectin",
    )

    initial_chain_assoc = relationship(NEBDatasetInitialMoleculeORM, cascade="all, delete-orphan", passive_deletes=True)

    __table_args__ = (
        Index("ix_neb_dataset_entry_dataset_id", "dataset_id"),
        Index("ix_neb_dataset_entry_name", "name"),
    )

    _qcportal_model_excludes = ["dataset_id"]


class NEBDatasetSpecificationORM(BaseORM):
    __tablename__ = "neb_dataset_specification"

    dataset_id = Column(Integer, ForeignKey("neb_dataset.id", ondelete="cascade"), primary_key=True)
    name = Column(String, primary_key=True)
    description = Column(String, nullable=True)
    specification_id = Column(Integer, ForeignKey(NEBSpecificationORM.id), nullable=False)

    specification = relationship(NEBSpecificationORM)

    __table_args__ = (
        Index("ix_neb_dataset_specification_dataset_id", "dataset_id"),
        Index("ix_neb_dataset_specification_name", "name"),
        Index("ix_neb_dataset_specification_specification_id", "specification_id"),
    )

    _qcportal_model_excludes = ["dataset_id", "specification_id"]


class NEBDatasetRecordItemORM(BaseORM):
    __tablename__ = "neb_dataset_record"

    dataset_id = Column(Integer, ForeignKey("neb_dataset.id", ondelete="cascade"), primary_key=True)
    entry_name = Column(String, primary_key=True)
    specification_name = Column(String, primary_key=True)
    record_id = Column(Integer, ForeignKey(NEBRecordORM.id), nullable=False)

    record = relationship(NEBRecordORM)

    __table_args__ = (
        ForeignKeyConstraint(
            ["dataset_id", "entry_name"],
            ["neb_dataset_entry.dataset_id", "neb_dataset_entry.name"],
            ondelete="cascade",
            onupdate="cascade",
        ),
        ForeignKeyConstraint(
            ["dataset_id", "specification_name"],
            ["neb_dataset_specification.dataset_id", "neb_dataset_specification.name"],
            ondelete="cascade",
            onupdate="cascade",
        ),
        Index("ix_neb_dataset_record_record_id", "record_id"),
    )

    _qcportal_model_excludes = ["dataset_id"]


class NEBDatasetORM(BaseDatasetORM):
    __tablename__ = "neb_dataset"

    id = Column(Integer, ForeignKey(BaseDatasetORM.id, ondelete="cascade"), primary_key=True)

    __mapper_args__ = {
        "polymorphic_identity": "neb",
    }


dataset_records_select = [
    select(
        BaseDatasetORM.id.label("dataset_id"),
        NEBDatasetRecordItemORM.entry_name.label("entry_name"),
        NEBDatasetRecordItemORM.specification_name.label("specification_name"),
        NEBDatasetRecordItemORM.record_id.label("record_id"),
    )
    .join(NEBDatasetRecordItemORM, BaseDatasetORM.id == NEBDatasetRecordItemORM.dataset_id)
    .where(BaseDatasetORM.dataset_type == "neb")
]



from __future__ import annotations

import secrets
from typing import TYPE_CHECKING, Optional

from flask.sessions import SessionInterface, SecureCookieSession

from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from flask import Flask, Response, Request


# The SecureCookieSession tracks accesses and modifications for us
class QCFFlaskSession(SecureCookieSession):
    session_key: Optional[str] = None


class QCFFlaskSessionInterface(SessionInterface):
    """
    Interface for a database-backed user session store

    open_session will see if data exists in the database and use that. A check is made for expired data as well.

    If there is something to store in the database, a random session_key will be generated in save_session.
    """

    def __init__(self, app: Flask):
        if not hasattr(app, "extensions") or "storage_socket" not in app.extensions:
            raise RuntimeError("The QCFFlaskSessionInterface requires the storage_socket extension to be initialized")

        self._storage_socket = app.extensions["storage_socket"]

    def open_session(self, app: Flask, request: Request) -> QCFFlaskSession:
        """
        Retrieve the session from the database

        If no session exists, or if the session has expired, an empty one will be created.

        This function will not set the session_key. That way, empty data is not persisted to the database
        """
        cookie_name = self.get_cookie_name(app)

        # Get the session_key from the cookie
        session_key = request.cookies.get(cookie_name)

        # No session cookie set - use empty session
        if not session_key:
            return QCFFlaskSession()

        session_data, last_accessed = self._storage_socket.auth.load_user_session(session_key)

        # No session data
        if session_data is None:
            return QCFFlaskSession()  # New session if none exists

        ret = QCFFlaskSession(initial=session_data)
        ret.session_key = session_key

        # Set/update basic data
        ret["user_agent"] = request.headers.get("User-Agent")
        ret["ip_address"] = request.remote_addr

        # IMPORTANT - Has the session expired?
        # Normally, this would be handled by flask with an expired cookie.
        # But the user may forge the expiration time
        # Note that app.permanent_session_lifetime is a timedelta
        expire_time = last_accessed + app.permanent_session_lifetime
        if expire_time < now_at_utc():
            return QCFFlaskSession()  # Consider the session invalid and return a new one

        return ret

    def save_session(self, app: Flask, session_data: QCFFlaskSession, response: Response) -> None:
        """
        Save the session to the database

        If no session_key is given on the QCFFlaskSession object, a new one will be generated. Otherwise, existing
        data will be replaced in the database.

        This function will also request to delete cookies if needed.
        """

        # This code adapted from flask sessions:
        # https://github.com/pallets/flask/blob/f61172b8dd3f962d33f25c50b2f5405e90ceffa5/src/flask/sessions.py#L350

        # See if the session_key exists in the data. If so, we should reuse the session
        cookie_name = self.get_cookie_name(app)
        session_key = session_data.session_key  # may be None

        if not session_data:
            if session_data.modified:
                # Session was modified to be empty, so delete the cookie
                response.delete_cookie(cookie_name)
                response.vary.add("Cookie")

            # session_key should only be attached if there is something in the database
            # That is, only if there was data loaded from the database in open_session
            # This should prevent unnecessary database queries.
            if session_key:
                self._storage_socket.auth.delete_user_session(user_session_key=session_key)
        else:
            user_id = int(session_data["user_id"])

            if not session_key:
                # create a unique ID if needed
                session_key = secrets.token_urlsafe(32)
                assert len(session_key) > 36  # Paranoid

                # Store for later, so we know to reuse this session
                session_data.session_key = session_key

            # Update the session data in the database
            self._storage_socket.auth.save_user_session(user_id, session_key, session_data)

            # Set the cookie in the response
            # Same name & session id, but extend the lifetime
            response.set_cookie(
                cookie_name,
                session_key,
                domain=app.config["QCFRACTAL_CONFIG"].api.user_session_cookie_domain,
                max_age=app.permanent_session_lifetime,
                httponly=app.config["QCFRACTAL_CONFIG"].api.user_session_cookie_httponly,
                samesite=app.config["QCFRACTAL_CONFIG"].api.user_session_cookie_samesite,
                secure=app.config["QCFRACTAL_CONFIG"].api.user_session_cookie_secure,
                partitioned=app.config["QCFRACTAL_CONFIG"].api.user_session_cookie_partitioned,
            )
            response.vary.add("Cookie")



from typing import Optional

from flask import Blueprint, current_app, redirect, abort, send_from_directory

home_v1 = Blueprint("home", __name__)


@home_v1.route("/", methods=["GET"])
@home_v1.route("/<path:file_path>", methods=["GET"])
def homepage(file_path: Optional[str] = None):
    # If the root is accessed, serve the static homepage site or do a redirect
    # If a specific file is accessed, only try to serve it from the directory

    homepage_redirect_url = current_app.config["QCFRACTAL_CONFIG"].homepage_redirect_url
    homepage_dir = current_app.config["QCFRACTAL_CONFIG"].homepage_directory

    if homepage_dir is not None:
        if file_path is None:
            file_path = "index.html"

        print("SENDING", homepage_dir, file_path)
        return send_from_directory(homepage_dir, file_path)
    if homepage_redirect_url is not None and file_path is None:
        # This should probably only be a temporary redirect (code 302).
        # This would make it easier to change the redirect in the settings, since
        # browsers shouldn't cache it
        return redirect(homepage_redirect_url, 302)

    # No homepage dir/redirect specified
    return abort(404)



from __future__ import annotations

from sqlalchemy import select, Column, Integer, ForeignKey, String, ForeignKeyConstraint, Index
from sqlalchemy.dialects.postgresql import JSONB, array_agg
from sqlalchemy.orm import relationship, column_property

from qcfractal.components.dataset_db_models import BaseDatasetORM
from qcfractal.components.molecules.db_models import MoleculeORM
from qcfractal.components.torsiondrive.record_db_models import (
    TorsiondriveRecordORM,
    TorsiondriveSpecificationORM,
)
from qcfractal.db_socket import BaseORM


class TorsiondriveDatasetMoleculeORM(BaseORM):
    """
    Association table torsiondrive -> initial molecules
    """

    __tablename__ = "torsiondrive_dataset_molecule"

    dataset_id = Column(Integer, primary_key=True)
    entry_name = Column(String, primary_key=True)
    molecule_id = Column(Integer, ForeignKey(MoleculeORM.id), primary_key=True)

    __table_args__ = (
        Index("ix_torsiondrive_dataset_molecule_dataset_id", "dataset_id"),
        Index("ix_torsiondrive_dataset_molecule_entry_name", "entry_name"),
        Index("ix_torsiondrive_dataset_molecule_molecule_id", "molecule_id"),
        ForeignKeyConstraint(
            ["dataset_id", "entry_name"],
            ["torsiondrive_dataset_entry.dataset_id", "torsiondrive_dataset_entry.name"],
            ondelete="cascade",
            onupdate="cascade",
        ),
    )


class TorsiondriveDatasetEntryORM(BaseORM):
    __tablename__ = "torsiondrive_dataset_entry"

    dataset_id = Column(Integer, ForeignKey("torsiondrive_dataset.id", ondelete="cascade"), primary_key=True)

    name = Column(String, primary_key=True)
    comment = Column(String)

    additional_keywords = Column(JSONB, nullable=False)
    additional_optimization_keywords = Column(JSONB, nullable=False)
    attributes = Column(JSONB, nullable=False)

    # Mark as deferred. We generally don't want to load it
    initial_molecule_ids = column_property(
        select(array_agg(TorsiondriveDatasetMoleculeORM.molecule_id))
        .where(TorsiondriveDatasetMoleculeORM.dataset_id == dataset_id)
        .where(TorsiondriveDatasetMoleculeORM.entry_name == name)
        .scalar_subquery(),
        deferred=True,
    )

    initial_molecules = relationship(
        MoleculeORM, secondary=TorsiondriveDatasetMoleculeORM.__tablename__, lazy="selectin", viewonly=True
    )

    initial_molecules_assoc = relationship(
        TorsiondriveDatasetMoleculeORM, cascade="all, delete-orphan", passive_deletes=True
    )

    __table_args__ = (
        Index("ix_torsiondrive_dataset_entry_dataset_id", "dataset_id"),
        Index("ix_torsiondrive_dataset_entry_name", "name"),
    )

    _qcportal_model_excludes = ["dataset_id", "initial_molecules_assoc", "initial_molecules_id"]


class TorsiondriveDatasetSpecificationORM(BaseORM):
    __tablename__ = "torsiondrive_dataset_specification"

    dataset_id = Column(Integer, ForeignKey("torsiondrive_dataset.id", ondelete="cascade"), primary_key=True)
    name = Column(String, primary_key=True)
    description = Column(String, nullable=True)
    specification_id = Column(Integer, ForeignKey(TorsiondriveSpecificationORM.id), nullable=False)

    specification = relationship(TorsiondriveSpecificationORM)

    __table_args__ = (
        Index("ix_torsiondrive_dataset_specification_dataset_id", "dataset_id"),
        Index("ix_torsiondrive_dataset_specification_name", "name"),
        Index("ix_torsiondrive_dataset_specification_specification_id", "specification_id"),
    )

    _qcportal_model_excludes = ["dataset_id", "specification_id"]


class TorsiondriveDatasetRecordItemORM(BaseORM):
    __tablename__ = "torsiondrive_dataset_record"

    dataset_id = Column(Integer, ForeignKey("torsiondrive_dataset.id", ondelete="cascade"), primary_key=True)
    entry_name = Column(String, primary_key=True)
    specification_name = Column(String, primary_key=True)
    record_id = Column(Integer, ForeignKey(TorsiondriveRecordORM.id), nullable=False)

    record = relationship(TorsiondriveRecordORM)

    __table_args__ = (
        ForeignKeyConstraint(
            ["dataset_id", "entry_name"],
            ["torsiondrive_dataset_entry.dataset_id", "torsiondrive_dataset_entry.name"],
            ondelete="cascade",
            onupdate="cascade",
        ),
        ForeignKeyConstraint(
            ["dataset_id", "specification_name"],
            ["torsiondrive_dataset_specification.dataset_id", "torsiondrive_dataset_specification.name"],
            ondelete="cascade",
            onupdate="cascade",
        ),
        Index("ix_torsiondrive_dataset_record_record_id", "record_id"),
    )

    _qcportal_model_excludes = ["dataset_id"]


class TorsiondriveDatasetORM(BaseDatasetORM):
    __tablename__ = "torsiondrive_dataset"

    id = Column(Integer, ForeignKey(BaseDatasetORM.id, ondelete="cascade"), primary_key=True)

    __mapper_args__ = {
        "polymorphic_identity": "torsiondrive",
    }


dataset_records_select = [
    select(
        BaseDatasetORM.id.label("dataset_id"),
        TorsiondriveDatasetRecordItemORM.entry_name.label("entry_name"),
        TorsiondriveDatasetRecordItemORM.specification_name.label("specification_name"),
        TorsiondriveDatasetRecordItemORM.record_id.label("record_id"),
    )
    .join(TorsiondriveDatasetRecordItemORM, BaseDatasetORM.id == TorsiondriveDatasetRecordItemORM.dataset_id)
    .where(BaseDatasetORM.dataset_type == "torsiondrive")
]



from __future__ import annotations

from typing import TYPE_CHECKING, Tuple, Optional, Union, Dict, Any

try:
    import pydantic.v1 as pydantic
except ImportError:
    import pydantic
from qcarchivetesting.helpers import read_procedure_data, read_record_data
from qcelemental.models import (
    Molecule,
    FailedOperation,
    ComputeError,
    AtomicResult as QCEl_AtomicResult,
)

from qcfractal.components.singlepoint.record_db_models import SinglepointRecordORM
from qcfractalcompute.compress import compress_result
from qcportal.record_models import PriorityEnum, RecordStatusEnum
from qcportal.singlepoint import SinglepointProtocols, QCSpecification, SinglepointDriver, SinglepointRecord
from qcportal.utils import now_at_utc
from qcarchivetesting.helpers import find_test_data

if TYPE_CHECKING:
    from qcfractal.db_socket import SQLAlchemySocket
    from qcportal.managers import ManagerName

all_test_data = find_test_data("sp_*")
all_includes = ["molecule", "wavefunction"]

test_specs = [
    QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis="6-31G*",
        keywords={"k": "value"},
        protocols=SinglepointProtocols(wavefunction="all"),
    ),
    QCSpecification(
        program="Prog2",
        driver=SinglepointDriver.gradient,
        method="Hf",
        basis="def2-TZVP",
        keywords={"k": "v"},
    ),
    QCSpecification(
        program="Prog3",
        driver=SinglepointDriver.hessian,
        method="pbe0",
        basis="",
        keywords={"o": 1, "v": 2.123},
        protocols=SinglepointProtocols(stdout=False, wavefunction="orbitals_and_eigenvalues"),
    ),
    QCSpecification(
        program="ProG4",
        driver=SinglepointDriver.hessian,
        method="pbe",
        basis=None,
        protocols=SinglepointProtocols(stdout=False, wavefunction="return_results"),
    ),
]


def load_procedure_data(name: str) -> Tuple[QCSpecification, Molecule, QCEl_AtomicResult]:
    data = read_procedure_data(name)

    return (
        pydantic.parse_obj_as(QCSpecification, data["specification"]),
        pydantic.parse_obj_as(Molecule, data["molecule"]),
        pydantic.parse_obj_as(QCEl_AtomicResult, data["result"]),
    )


def load_error_procedure_data(name: str) -> Tuple[QCSpecification, Molecule, FailedOperation]:
    data = read_procedure_data(name)
    return (
        pydantic.parse_obj_as(QCSpecification, data["specification"]),
        pydantic.parse_obj_as(Molecule, data["molecule"]),
        pydantic.parse_obj_as(FailedOperation, data["result"]),
    )


def load_record_data(name: str) -> SinglepointRecord:
    data = read_record_data(name)
    return SinglepointRecord(**data)

def load_error_record_data(name: str) -> FailedOperation:
    data = read_record_data(name)
    return FailedOperation(**data)


def submit_procedure_data(
    storage_socket: SQLAlchemySocket,
    name: str,
    compute_tag: Optional[str] = "*",
    compute_priority: PriorityEnum = PriorityEnum.normal,
    find_existing: bool = True,
) -> Tuple[int, QCEl_AtomicResult]:
    input_spec, molecule, result = load_procedure_data(name)
    meta, record_ids = storage_socket.records.singlepoint.add(
        [molecule], input_spec, compute_tag, compute_priority, None, find_existing
    )
    assert meta.success
    assert len(record_ids) == 1
    assert meta.n_inserted == 1

    return record_ids[0], result


def run_procedure_data(
    storage_socket: SQLAlchemySocket,
    manager_name: ManagerName,
    name: str,
    compute_tag: Optional[str] = "*",
    compute_priority: PriorityEnum = PriorityEnum.normal,
    end_status: RecordStatusEnum = RecordStatusEnum.complete,
):
    time_0 = now_at_utc()
    record_id, result = submit_procedure_data(storage_socket, name, compute_tag, compute_priority)
    time_1 = now_at_utc()

    with storage_socket.session_scope() as session:
        record = session.get(SinglepointRecordORM, record_id)
        assert record.status == RecordStatusEnum.waiting

    if end_status == RecordStatusEnum.error:
        result = FailedOperation(
            error=ComputeError(error_type="test_error", error_message="this is just a test error"),
        )

    manager_programs = storage_socket.managers.get([manager_name.fullname])[0]["programs"]
    tasks = storage_socket.tasks.claim_tasks(manager_name.fullname, manager_programs, [compute_tag], limit=100)
    assert len(tasks) == 1

    result_compressed = compress_result(result.dict())
    result_dict = {tasks[0]["id"]: result_compressed}

    storage_socket.tasks.update_finished(manager_name.fullname, result_dict)
    time_2 = now_at_utc()

    with storage_socket.session_scope() as session:
        record = session.get(SinglepointRecordORM, record_id)
        assert record.status == end_status
        assert time_0 < record.created_on < time_1
        assert time_1 < record.modified_on < time_2
        assert time_1 < record.compute_history[0].modified_on < time_2

    return record_id


def compare_singlepoint_specs(
    input_spec: Union[QCSpecification, Dict[str, Any]],
    output_spec: Union[QCSpecification, Dict[str, Any]],
) -> bool:
    if isinstance(input_spec, dict):
        input_spec = QCSpecification(**input_spec)
    if isinstance(output_spec, dict):
        output_spec = QCSpecification(**output_spec)

    return input_spec == output_spec



from __future__ import annotations

import itertools
from typing import TYPE_CHECKING, Optional

import pytest

from qcarchivetesting import load_molecule_data
from qcfractal.components.torsiondrive.record_db_models import TorsiondriveRecordORM
from qcportal.optimization import OptimizationSpecification
from qcportal.record_models import RecordStatusEnum, PriorityEnum
from qcportal.singlepoint import QCSpecification
from qcportal.torsiondrive import TorsiondriveKeywords, TorsiondriveSpecification
from qcportal.utils import now_at_utc
from .testing_helpers import compare_torsiondrive_specs, test_specs, submit_procedure_data, run_procedure_data

if TYPE_CHECKING:
    from qcarchivetesting.testing_classes import QCATestingSnowflake
    from qcportal import PortalClient


def test_torsiondrive_client_tag_priority_as_service(snowflake_client: PortalClient):
    peroxide2 = load_molecule_data("peroxide2")

    for tag, priority in itertools.product(["*", "tag99"], list(PriorityEnum)):
        meta1, id1 = snowflake_client.add_torsiondrives(
            [[peroxide2]],
            "torsiondrive",
            optimization_specification=OptimizationSpecification(
                program="geometric",
                qc_specification=QCSpecification(
                    program="psi4",
                    method="hf",
                    basis="sto-3g",
                    driver="deferred",
                    keywords={"tag_priority": [tag, priority]},
                ),
            ),
            keywords=TorsiondriveKeywords(dihedrals=[(1, 2, 3, 4)], grid_spacing=[15], energy_upper_limit=0.04),
            compute_priority=priority,
            compute_tag=tag,
        )

        assert meta1.n_inserted == 1
        rec = snowflake_client.get_records(id1, include=["service"])
        assert rec[0].service.compute_tag == tag
        assert rec[0].service.compute_priority == priority


@pytest.mark.parametrize("spec", test_specs)
def test_torsiondrive_client_add_get(
    submitter_client: PortalClient,
    spec: TorsiondriveSpecification,
):
    hooh = load_molecule_data("peroxide2")
    td_mol_1 = load_molecule_data("td_C9H11NO2_1")
    td_mol_2 = load_molecule_data("td_C9H11NO2_2")

    time_0 = now_at_utc()
    meta, id = submitter_client.add_torsiondrives(
        [[hooh], [td_mol_1, td_mol_2]],
        "torsiondrive",
        keywords=spec.keywords,
        optimization_specification=spec.optimization_specification,
        compute_tag="tag1",
        compute_priority=PriorityEnum.low,
    )
    time_1 = now_at_utc()
    assert meta.success

    recs = submitter_client.get_torsiondrives(id, include=["service", "initial_molecules"])
    assert len(recs) == 2

    for r in recs:
        assert r.record_type == "torsiondrive"
        assert r.record_type == "torsiondrive"
        assert compare_torsiondrive_specs(spec, r.specification)

        assert r.status == RecordStatusEnum.waiting
        assert r.children_status == {}

        assert r.service.compute_tag == "tag1"
        assert r.service.compute_priority == PriorityEnum.low

        assert r.creator_user == submitter_client.username

        assert time_0 < r.created_on < time_1
        assert time_0 < r.modified_on < time_1

    assert len(recs[0].initial_molecules) == 1
    assert len(recs[1].initial_molecules) == 2

    assert recs[0].initial_molecules[0].get_hash() == hooh.get_hash()

    # Not necessarily in the input order
    hash1 = recs[1].initial_molecules[0].get_hash()
    hash2 = recs[1].initial_molecules[1].get_hash()
    assert {hash1, hash2} == {td_mol_1.get_hash(), td_mol_2.get_hash()}


@pytest.mark.parametrize("spec", test_specs)
@pytest.mark.parametrize("find_existing", [True, False])
def test_torsiondrive_client_add_duplicate(
    submitter_client: PortalClient,
    spec: TorsiondriveSpecification,
    find_existing: bool,
):
    hooh = load_molecule_data("peroxide2")
    td_mol_1 = load_molecule_data("td_C9H11NO2_1")
    td_mol_2 = load_molecule_data("td_C9H11NO2_2")

    all_mols = [[hooh], [td_mol_1, td_mol_2]]

    meta, id = submitter_client.add_torsiondrives(
        all_mols,
        "torsiondrive",
        keywords=spec.keywords,
        optimization_specification=spec.optimization_specification,
        compute_tag="tag1",
        compute_priority=PriorityEnum.low,
        find_existing=True,
    )

    assert meta.success
    assert meta.n_inserted == len(all_mols)

    meta, id2 = submitter_client.add_torsiondrives(
        all_mols,
        "torsiondrive",
        keywords=spec.keywords,
        optimization_specification=spec.optimization_specification,
        compute_tag="tag1",
        compute_priority=PriorityEnum.low,
        find_existing=find_existing,
    )

    if find_existing:
        assert meta.n_existing == len(all_mols)
        assert meta.n_inserted == 0
        assert id == id2
    else:
        assert meta.n_existing == 0
        assert meta.n_inserted == len(all_mols)
        assert set(id).isdisjoint(id2)


def test_torsiondrive_client_add_existing_molecule(snowflake_client: PortalClient):
    spec = test_specs[0]

    mol1 = load_molecule_data("td_C9H11NO2_1")
    mol2 = load_molecule_data("td_C9H11NO2_2")

    # Add a molecule separately
    _, mol_ids = snowflake_client.add_molecules([mol2])

    # Now add records
    meta, id = snowflake_client.add_torsiondrives(
        [[mol1, mol2], [mol2, mol1]],
        "torsiondrive",
        keywords=spec.keywords,
        optimization_specification=spec.optimization_specification,
        compute_tag="tag1",
        compute_priority=PriorityEnum.low,
    )

    assert meta.success
    assert meta.n_inserted == 1
    assert meta.n_existing == 1

    recs = snowflake_client.get_torsiondrives(id, include=["initial_molecules"])
    assert len(recs) == 2
    assert recs[0].id == recs[1].id

    rec_mols = {x.id for x in recs[0].initial_molecules}
    _, mol_ids_2 = snowflake_client.add_molecules([mol1])
    assert rec_mols == set(mol_ids + mol_ids_2)


def test_torsiondrive_client_delete(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    td_id = run_procedure_data(storage_socket, activated_manager_name, "td_H2O2_mopac_pm6")

    with storage_socket.session_scope() as session:
        rec = session.get(TorsiondriveRecordORM, td_id)
        child_ids = [x.optimization_id for x in rec.optimizations]

    meta = snowflake_client.delete_records(td_id, soft_delete=True, delete_children=False)
    assert meta.success
    assert meta.deleted_idx == [0]
    assert meta.n_children_deleted == 0

    child_recs = snowflake_client.get_records(child_ids, missing_ok=True)
    assert all(x.status == RecordStatusEnum.complete for x in child_recs)
    td_rec = snowflake_client.get_records(td_id)
    assert td_rec.children_status == {RecordStatusEnum.complete: len(child_ids)}

    snowflake_client.undelete_records(td_id)

    meta = snowflake_client.delete_records(td_id, soft_delete=True, delete_children=True)
    assert meta.success
    assert meta.deleted_idx == [0]
    assert meta.n_children_deleted == len(child_ids)

    child_recs = snowflake_client.get_records(child_ids, missing_ok=True)
    assert all(x.status == RecordStatusEnum.deleted for x in child_recs)
    td_rec = snowflake_client.get_records(td_id)
    assert td_rec.children_status == {RecordStatusEnum.deleted: len(child_ids)}

    meta = snowflake_client.delete_records(td_id, soft_delete=False, delete_children=True)
    assert meta.success
    assert meta.deleted_idx == [0]
    assert meta.n_children_deleted == len(child_ids)

    recs = snowflake_client.get_torsiondrives(td_id, missing_ok=True)
    assert recs is None

    child_recs = snowflake_client.get_records(child_ids, missing_ok=True)
    assert all(x is None for x in child_recs)

    # DB should be pretty empty now
    query_res = snowflake_client.query_records()
    query_res_l = list(query_res)
    assert len(query_res_l) == 0


def test_torsiondrive_client_harddelete_nochildren(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    td_id = run_procedure_data(storage_socket, activated_manager_name, "td_H2O2_mopac_pm6")

    with storage_socket.session_scope() as session:
        rec = session.get(TorsiondriveRecordORM, td_id)
        child_ids = [x.optimization_id for x in rec.optimizations]

    meta = snowflake_client.delete_records(td_id, soft_delete=False, delete_children=False)
    assert meta.success
    assert meta.deleted_idx == [0]
    assert meta.n_children_deleted == 0

    recs = snowflake_client.get_torsiondrives(td_id, missing_ok=True)
    assert recs is None

    child_recs = snowflake_client.get_records(child_ids, missing_ok=True)
    assert all(x is not None for x in child_recs)


def test_torsiondrive_client_delete_opt_inuse(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    td_id = run_procedure_data(storage_socket, activated_manager_name, "td_H2O2_mopac_pm6")

    with storage_socket.session_scope() as session:
        rec = session.get(TorsiondriveRecordORM, td_id)
        child_ids = [x.optimization_id for x in rec.optimizations]

    meta = snowflake_client.delete_records(child_ids[0], soft_delete=False)
    assert meta.success is False
    assert meta.error_idx == [0]

    ch_rec = snowflake_client.get_records(child_ids[0])
    assert ch_rec is not None


def test_torsiondrive_client_query(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    id_1, _ = submit_procedure_data(storage_socket, "td_H2O2_mopac_pm6")
    id_2, _ = submit_procedure_data(storage_socket, "td_H2O2_psi4_pbe")
    id_3, _ = submit_procedure_data(storage_socket, "td_C9H11NO2_mopac_pm6")
    id_4, _ = submit_procedure_data(storage_socket, "td_H2O2_psi4_pbe0")

    all_tds = snowflake_client.get_torsiondrives([id_1, id_2, id_3, id_4], include=["initial_molecules"])
    mol_ids = [x.initial_molecules[0].id for x in all_tds]

    query_res = snowflake_client.query_torsiondrives(qc_program=["psi4"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 2

    query_res = snowflake_client.query_torsiondrives(qc_program=["nothing"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    query_res = snowflake_client.query_torsiondrives(initial_molecule_id=[mol_ids[0], 9999])
    query_res_l = list(query_res)
    assert len(query_res_l) == 3

    # query for optimization program
    query_res = snowflake_client.query_torsiondrives(optimization_program=["geometric"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 4

    # query for optimization program
    query_res = snowflake_client.query_torsiondrives(optimization_program=["geometric123"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    # query for basis
    query_res = snowflake_client.query_torsiondrives(qc_basis=["sTO-3g"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 2

    query_res = snowflake_client.query_torsiondrives(qc_basis=[None])
    query_res_l = list(query_res)
    assert len(query_res_l) == 2

    query_res = snowflake_client.query_torsiondrives(qc_basis=[""])
    query_res_l = list(query_res)
    assert len(query_res_l) == 2

    # query for method
    query_res = snowflake_client.query_torsiondrives(qc_method=["pm6"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 2

    # Query by default returns everything
    query_res = snowflake_client.query_torsiondrives()
    query_res_l = list(query_res)
    assert len(query_res_l) == 4

    # Query by default (with a limit)
    query_res = snowflake_client.query_torsiondrives(limit=1)
    query_res_l = list(query_res)
    assert len(query_res_l) == 1



"""Use computed column for dataset lname

Revision ID: 13cb230def11
Revises: d1ee87a66b71
Create Date: 2023-09-14 18:50:20.067940

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "13cb230def11"
down_revision = "d1ee87a66b71"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute(sa.text("ALTER TABLE base_dataset DROP COLUMN lname;"))
    op.execute(
        sa.text(
            "ALTER TABLE base_dataset ADD COLUMN lname CHARACTER VARYING(100) NOT NULL GENERATED ALWAYS AS (lower((name)::text)) stored;"
        )
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute(sa.text("ALTER TABLE base_dataset DROP COLUMN lname;"))
    op.execute(sa.text("ALTER TABLE base_dataset ADD COLUMN lname CHARACTER VARYING(100)"))
    op.execute(sa.text("UPDATE base_dataset SET lname = lower(name)"))
    op.alter_column("base_dataset", "lname", nullable=False)
    op.create_unique_constraint("ux_base_dataset_dataset_type_lname", "base_dataset", ["dataset_type", "lname"])

    # ### end Alembic commands ###



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from qcfractal.components.singlepoint.testing_helpers import load_procedure_data, run_procedure_data
from qcportal import PortalRequestError
from qcportal.molecules import Molecule
from qcportal.record_models import PriorityEnum
from qcportal.singlepoint import SinglepointDataset

if TYPE_CHECKING:
    from qcportal import PortalClient
    from qcarchivetesting.testing_classes import QCATestingSnowflake


@pytest.mark.parametrize(
    "dataset_type", ["singlepoint", "optimization", "torsiondrive", "gridoptimization", "manybody", "reaction", "neb"]
)
def test_dataset_client_add_get(submitter_client: PortalClient, dataset_type: str):
    ds = submitter_client.add_dataset(
        dataset_type,
        "Test dataset",
        "Test Description",
        "a Tagline",
        ["tag1", "tag2"],
        "new_group",
        {"prov_key_1": "prov_value_1"},
        True,
        "def_tag",
        PriorityEnum.low,
        {"meta_key_1": "meta_value_1"},
    )

    assert ds.dataset_type == dataset_type
    assert ds.name == "Test dataset"
    assert ds.description == "Test Description"
    assert ds.tagline == "a Tagline"
    assert ds.tags == ["tag1", "tag2"]
    assert ds.provenance == {"prov_key_1": "prov_value_1"}
    assert ds.default_compute_tag == "def_tag"
    assert ds.default_compute_priority == PriorityEnum.low
    assert ds.extras == {"meta_key_1": "meta_value_1"}

    assert ds.creator_user == submitter_client.username

    # case insensitive
    ds2 = submitter_client.get_dataset(dataset_type, "test DATASET")
    assert ds2.id == ds.id


def test_dataset_client_status(snowflake: QCATestingSnowflake):
    snowflake_client = snowflake.client()
    storage_socket = snowflake.get_storage_socket()
    manager_name, _ = snowflake.activate_manager()

    ds: SinglepointDataset = snowflake_client.add_dataset("singlepoint", "Test dataset")
    assert ds.status() == {}

    input_spec, molecule, _ = load_procedure_data("sp_psi4_peroxide_energy_wfn")
    run_procedure_data(storage_socket, manager_name, "sp_psi4_peroxide_energy_wfn")

    molecule_2 = Molecule(symbols=["b"], geometry=[0, 0, 0])

    # Add this as a part of the dataset
    ds.add_specification("spec_1", input_spec)
    ds.add_entry(name="test_molecule", molecule=molecule)
    ds.add_entry(name="test_molecule_2", molecule=molecule_2)
    ds.submit()

    stat1 = ds.status()
    stat2 = snowflake_client.get_dataset_status_by_id(ds.id)
    assert stat1 == stat2


def test_dataset_client_status_by_tag(snowflake: QCATestingSnowflake):
    snowflake_client = snowflake.client()
    manager_name, _ = snowflake.activate_manager()

    ds: SinglepointDataset = snowflake_client.add_dataset("singlepoint", "Test dataset")
    assert ds.status_by_compute_tag() == []

    input_spec, molecule, _ = load_procedure_data("sp_psi4_peroxide_energy_wfn")

    molecule_2 = Molecule(symbols=["b"], geometry=[0, 0, 0])

    # Add this as a part of the dataset
    ds.add_specification("spec_1", input_spec)
    ds.add_entry(name="test_molecule", molecule=molecule)
    ds.add_entry(name="test_molecule_2", molecule=molecule_2)
    ds.submit(entry_names=["test_molecule"], compute_tag="tag1")
    ds.submit(entry_names=["test_molecule_2"], compute_tag="tag2")

    status = ds.status_by_compute_tag()
    assert len(status) == 2
    assert {s[0] for s in status} == {"tag1", "tag2"}
    assert all(s[1] == "waiting" for s in status)
    assert all(s[2] == 1 for s in status)


def test_dataset_client_add_same_name(snowflake_client: PortalClient):
    ds1 = snowflake_client.add_dataset("singlepoint", "Test dataset")
    ds2 = snowflake_client.add_dataset("optimization", "Test dataset")

    assert ds1.id != ds2.id


def test_dataset_client_add_duplicate(snowflake_client: PortalClient):
    ds = snowflake_client.add_dataset("singlepoint", "Test dataset")

    with pytest.raises(PortalRequestError, match=r"Dataset.*already exists"):
        # existing_ok = False by default
        snowflake_client.add_dataset("singlepoint", "TEST DATASET")

    ds2 = snowflake_client.add_dataset("singlepoint", "Test dataset", existing_ok=True)
    assert ds.id == ds2.id


def test_dataset_client_delete_empty(snowflake_client: PortalClient):
    ds = snowflake_client.add_dataset("singlepoint", "Test dataset")
    ds_id = ds.id

    ds = snowflake_client.get_dataset("singlepoint", "Test dataset")
    assert ds.id == ds_id

    snowflake_client.delete_dataset(ds_id, False)

    with pytest.raises(PortalRequestError, match=r"Could not find singlepoint dataset with name"):
        snowflake_client.get_dataset("singlepoint", "Test dataset")

    with pytest.raises(PortalRequestError, match=r"Could not find dataset with id"):
        snowflake_client.get_dataset_by_id(ds_id)

    all_ds = snowflake_client.list_datasets()
    len(all_ds) == 0


def test_dataset_client_query_dataset_records(snowflake: QCATestingSnowflake):
    snowflake_client = snowflake.client()
    storage_socket = snowflake.get_storage_socket()
    manager_name, _ = snowflake.activate_manager()

    ds: SinglepointDataset = snowflake_client.add_dataset("singlepoint", "Test dataset")
    assert ds.status() == {}

    input_spec, molecule, _ = load_procedure_data("sp_psi4_peroxide_energy_wfn")
    run_procedure_data(storage_socket, manager_name, "sp_psi4_peroxide_energy_wfn")

    molecule_2 = Molecule(symbols=["b"], geometry=[0, 0, 0])

    # Add this as a part of the dataset
    ds.add_specification("spec_1", input_spec)
    ds.add_entry(name="test_molecule", molecule=molecule)
    ds.add_entry(name="test_molecule_2", molecule=molecule_2)
    ds.submit()

    # Query records belonging to a dataset
    rec_id_2 = ds.get_record("test_molecule_2", "spec_1").id
    mol_id_2 = ds.get_entry("test_molecule_2").molecule.id

    query_res = snowflake_client.query_singlepoints(dataset_id=ds.id)
    query_res_l = list(query_res)
    assert len(query_res_l) == 2

    query_res = snowflake_client.query_singlepoints(dataset_id=ds.id, molecule_id=mol_id_2)
    query_res_l = list(query_res)
    assert len(query_res_l) == 1
    assert query_res_l[0].id == rec_id_2

    query_res = snowflake_client.query_singlepoints(dataset_id=ds.id + 1, molecule_id=mol_id_2)
    assert query_res._fetched == 0

    # Query which dataset contains a record
    rec_info = snowflake_client.query_dataset_records([rec_id_2])
    assert len(rec_info) == 1
    assert rec_info[0]["dataset_id"] == 1
    assert rec_info[0]["entry_name"] == "test_molecule_2"

    # Query which dataset contains a record
    ds.remove_records(entry_names="test_molecule_2", specification_names="spec_1", delete_records=True)
    rec_info = snowflake_client.query_dataset_records([rec_id_2])
    assert len(rec_info) == 0


def test_dataset_rename(snowflake_client: PortalClient):
    ds = snowflake_client.add_dataset("singlepoint", "Test dataset")
    assert ds.name == "Test dataset"
    ds_id = ds.id
    assert ds.status() == {}

    ds.set_name("Different name")
    ds = snowflake_client.get_dataset_by_id(ds_id)
    assert ds.name == "Different name"

    ds.set_name("different name")
    ds = snowflake_client.get_dataset_by_id(ds_id)
    assert ds.name == "different name"


def test_dataset_rename_specifications(snowflake_client: PortalClient):
    ds: SinglepointDataset = snowflake_client.add_dataset("singlepoint", "Test dataset")
    assert ds.status() == {}

    input_spec, molecule, _ = load_procedure_data("sp_psi4_peroxide_energy_wfn")
    ds.add_specification("spec_1", input_spec)

    assert len(ds.specifications) == 1

    ds = snowflake_client.get_dataset_by_id(ds.id)
    assert len(ds.specifications) == 1
    ds.rename_specification("spec_1", "spec_2")

    assert "spec_2" in ds.specifications
    assert "spec_1" not in ds.specifications

    ds = snowflake_client.get_dataset_by_id(ds.id)
    assert "spec_2" in ds.specifications
    assert "spec_1" not in ds.specifications

    # Rename without actually renaming
    ds.rename_specification("spec_2", "spec_2")
    assert "spec_2" in ds.specifications
    assert "spec_1" not in ds.specifications

    ds = snowflake_client.get_dataset_by_id(ds.id)
    assert "spec_2" in ds.specifications
    assert "spec_1" not in ds.specifications


def test_dataset_rename_entries(snowflake_client: PortalClient):
    ds: SinglepointDataset = snowflake_client.add_dataset("singlepoint", "Test dataset")
    assert ds.status() == {}

    input_spec, molecule, _ = load_procedure_data("sp_psi4_peroxide_energy_wfn")
    molecule_2 = Molecule(symbols=["b"], geometry=[0, 0, 0])

    ds.add_entry(name="test_molecule", molecule=molecule)
    ds.add_entry(name="test_molecule_2", molecule=molecule_2)

    assert len(ds.entry_names) == 2

    ds = snowflake_client.get_dataset_by_id(ds.id)
    assert len(ds.entry_names) == 2
    ds.rename_entries({"test_molecule": "different_name"})

    assert "different_name" in ds.entry_names
    assert "test_molecule" not in ds.entry_names
    assert "test_molecule_2" in ds.entry_names

    ds.rename_entries({"test_molecule_2": "test_molecule_2"})

    ds = snowflake_client.get_dataset_by_id(ds.id)
    assert len(ds.entry_names) == 2

    assert "different_name" in ds.entry_names
    assert "test_molecule" not in ds.entry_names
    assert "test_molecule_2" in ds.entry_names


def test_dataset_client_get_computed_properties(snowflake: QCATestingSnowflake):
    snowflake_client = snowflake.client()
    storage_socket = snowflake.get_storage_socket()
    manager_name, _ = snowflake.activate_manager()

    ds: SinglepointDataset = snowflake_client.add_dataset("singlepoint", "Test dataset")
    assert ds.status() == {}

    input_spec, molecule, _ = load_procedure_data("sp_psi4_peroxide_energy_wfn")
    record_id = run_procedure_data(storage_socket, manager_name, "sp_psi4_peroxide_energy_wfn")

    # Add this as a part of the dataset
    ds.add_specification("spec_1", input_spec)
    ds.add_entry(name="test_molecule", molecule=molecule)
    ds.submit()

    assert ds.status()["spec_1"]["complete"] == 1

    computed_prop = ds.computed_properties
    assert "spec_1" in computed_prop
    assert "scf_total_energy" in computed_prop["spec_1"]
    assert "calcinfo_natom" in computed_prop["spec_1"]


def test_dataset_client_copy_from_incompatible(snowflake_client: PortalClient):
    ds_1 = snowflake_client.add_dataset("singlepoint", "Test sp dataset")
    ds_2 = snowflake_client.add_dataset("optimization", "Test opt dataset")

    with pytest.raises(PortalRequestError, match="does not match destination type"):
        ds_2.copy_records_from(ds_1.id)



from qcarchivetesting import load_hash_test_data
from qcfractal.components.singlepoint.record_db_models import QCSpecificationORM
from qcfractal.components.testing_fixtures import spec_test_runner
from qcfractal.db_socket import SQLAlchemySocket
from qcportal.singlepoint import QCSpecification, SinglepointDriver, SinglepointProtocols


def test_singlepoint_hash_canaries(storage_socket: SQLAlchemySocket):
    # Test data is hash : spec dict
    test_data = load_hash_test_data("qc_specification_tests")
    spec_map = [(k, QCSpecification(**v)) for k, v in test_data.items()]

    specs = [x[1] for x in spec_map]
    meta, ids = storage_socket.records.singlepoint.add_specifications(specs)
    assert meta.success
    assert len(ids) == len(specs)
    assert meta.n_existing == 0

    with storage_socket.session_scope() as session:
        for spec_id, (spec_hash, _) in zip(ids, spec_map):
            spec_orm = session.get(QCSpecificationORM, spec_id)
            assert spec_orm.specification_hash == spec_hash


def test_singlepoint_socket_add_specification_same_1(spec_test_runner):
    spec1 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis="6-31G*",
        keywords={"k": "value"},
        protocols=SinglepointProtocols(wavefunction="all"),
    )

    spec_test_runner("singlepoint", spec1, spec1, True)


def test_singlepoint_socket_add_specification_same_2(spec_test_runner):
    # Test case sensitivity
    spec1 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis="6-31G*",
        keywords={"k": "value"},
        protocols=SinglepointProtocols(wavefunction="all"),
    )

    spec2 = QCSpecification(
        program="Prog1",
        driver=SinglepointDriver.energy,
        method="b3LYP",
        basis="6-31g*",
        keywords={"k": "value"},
        protocols=SinglepointProtocols(wavefunction="all"),
    )

    spec_test_runner("singlepoint", spec1, spec2, True)


def test_singlepoint_socket_add_specification_same_3(spec_test_runner):
    # Test keywords defaults
    spec1 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis="6-31G*",
        keywords={},
        protocols=SinglepointProtocols(wavefunction="orbitals_and_eigenvalues"),
    )

    spec2 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis="6-31G*",
        protocols=SinglepointProtocols(wavefunction="orbitals_and_eigenvalues"),
    )

    spec_test_runner("singlepoint", spec1, spec2, True)


def test_singlepoint_socket_add_specification_same_4(spec_test_runner):
    # Test protocols defaults
    spec1 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis="6-31G*",
        protocols=SinglepointProtocols(),
    )

    spec2 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis="6-31G*",
    )

    spec_test_runner("singlepoint", spec1, spec2, True)


def test_singlepoint_socket_add_specification_same_5(spec_test_runner):
    # Test protocols defaults
    spec1 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis="6-31G*",
        protocols=SinglepointProtocols(
            wavefunction="none", stdout=True, error_correction={"default_policy": True, "policies": {}}
        ),
    )

    spec2 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis="6-31G*",
    )

    spec_test_runner("singlepoint", spec1, spec2, True)


def test_singlepoint_socket_add_specification_same_6(spec_test_runner):
    # Test protocols defaults
    spec1 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis="6-31G*",
        protocols=SinglepointProtocols(
            wavefunction="none", stdout=True, error_correction={"default_policy": True, "policies": None}
        ),
    )

    spec2 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis="6-31G*",
    )

    spec_test_runner("singlepoint", spec1, spec2, True)


def test_singlepoint_socket_add_specification_same_7(spec_test_runner):
    # Test basis none vs empty string
    spec1 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis=None,
        protocols=SinglepointProtocols(),
    )

    spec2 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis="",
        protocols=SinglepointProtocols(),
    )

    spec_test_runner("singlepoint", spec1, spec2, True)


def test_singlepoint_socket_add_specification_same_8(spec_test_runner):
    # keyword ordering
    spec1 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis="sto-3g",
        protocols=SinglepointProtocols(),
        keywords={"a": 10, "b": "str"},
    )

    spec2 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis="sto-3g",
        protocols=SinglepointProtocols(),
        keywords={"b": "str", "a": 10},
    )

    spec_test_runner("singlepoint", spec1, spec2, True)


def test_singlepoint_socket_add_specification_diff_1(spec_test_runner):
    # Test different programs
    spec1 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis="sto-3g",
    )

    spec2 = spec1.copy(update={"program": "prog2"})

    spec_test_runner("singlepoint", spec1, spec2, False)


def test_singlepoint_socket_add_specification_diff_2(spec_test_runner):
    # Test different driver
    spec1 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.gradient,
        method="b3lyp",
        basis=None,
    )

    spec2 = spec1.copy(update={"driver": SinglepointDriver.energy})

    spec_test_runner("singlepoint", spec1, spec2, False)


def test_singlepoint_socket_add_specification_diff_3(spec_test_runner):
    # Test different keywords
    spec1 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis=None,
    )

    spec2 = spec1.copy(update={"keywords": {"k": "value"}})

    spec_test_runner("singlepoint", spec1, spec2, False)


def test_singlepoint_socket_add_specification_diff_4(spec_test_runner):
    # Test different keywords
    spec1 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis="sto-3g",
        keywords={"k": 1.0e-8},
    )

    spec2 = spec1.copy(update={"keywords": {"k": 1.0e-9}})

    spec_test_runner("singlepoint", spec1, spec2, False)


def test_singlepoint_socket_add_specification_diff_5(spec_test_runner):
    # Test different keywords
    spec1 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis="sto-3g",
        keywords={"k": "value"},
    )

    spec2 = spec1.copy(update={"keywords": {"k": "value", "k2": "value2"}})

    spec_test_runner("singlepoint", spec1, spec2, False)


def test_singlepoint_socket_add_specification_diff_6(spec_test_runner):
    # Test different protocols
    spec1 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis=None,
        protocols=SinglepointProtocols(),
    )

    spec2 = spec1.copy(update={"protocols": SinglepointProtocols(stdout=False)})

    spec_test_runner("singlepoint", spec1, spec2, False)


def test_singlepoint_socket_add_specification_diff_7(spec_test_runner):
    # Test different protocols
    spec1 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis=None,
        protocols=SinglepointProtocols(),
    )

    spec2 = spec1.copy(update={"protocols": SinglepointProtocols(wavefunction="orbitals_and_eigenvalues")})

    spec_test_runner("singlepoint", spec1, spec2, False)
